"""
This file runs before tests are collected in order to set up the repository for tests.
"""

import re
from os import sep
from pathlib import Path

import pytest

# Delete all test output files before beginning so we don't pass a test based on a
# previous run's output.
for path in Path(".").rglob(f"**{sep}tests{sep}output{sep}*.txt"):
    path.unlink()
for path in Path(".").rglob(f"**{sep}tests{sep}output{sep}*.tmp"):
    path.unlink()
for path in Path(".").rglob(f"**{sep}tests{sep}output{sep}*.json"):
    path.unlink()


@pytest.fixture(scope="session")
def process_id(worker_id: str) -> int:  # pragma: no cover
    """
    Derive a process ID from a worker ID.
    """
    if worker_id == "master":
        return 0
    return int(re.sub("[^0-9]", "", worker_id))



"""
Types for applications.
"""

from typing import TYPE_CHECKING, Generic, TypedDict, TypeVar

if TYPE_CHECKING:  # pragma: no cover
    from pax25.services.connection.connection import Connection
    from pax25.station import Station


S = TypeVar("S")


class BaseApplication(Generic[S]):
    """
    Raw BaseApplication that implements the minimum contract that Connection and
    the FrameRouter expect. In most cases you will not want to inherit from this, but
    from Application instead.
    """

    name: str
    settings: S
    # If this is set, will be used when constructing the default station ID beacon
    # to advertise its availability on that interface. If blank, will be omitted.
    short_name: str = ""

    def __init__(
        self, *, name: str, station: "Station", settings: S
    ):  # pragma: no cover
        """
        Initialize a copy of this application. Usually you want to store
        the settings and whether this is being invoked as a proxy object.
        """
        raise NotImplementedError

    def on_connect(self, connection: "Connection") -> None:
        """
        Called when a new connection is established. For instance, if we need to store
        any state for the user's session, we'd set it here. You might also send a
        welcome message.
        """

    def on_disconnect(self, connection: "Connection") -> None:
        """
        Called when a connection is being disconnected. Perform any cleanup here.
        """

    def on_killed(self, connection: "Connection") -> None:
        """
        Run when the connection has been killed. None of the other hooks, such as
        on_connect, or on_shutdown are guaranteed to have run before this.
        You are unlikely to need this hook, but you might for some special applications.
        """

    def on_bytes(self, connection: "Connection", bytes_received: bytes) -> None:
        """
        Called when bytes are received from a connection for this application.
        """


class BasicApplicationState(TypedDict):
    """
    State for the Application.
    """

    command: bytes


BasicStateTable = dict["Connection", BasicApplicationState]



"""
Types for the Personal Bulletin Board System.
"""

from dataclasses import dataclass
from typing import Literal, TypedDict

from pax25.services.connection.connection import Connection
from pax25.types import Version


class Message(TypedDict):
    """
    Message format for board posts.
    """

    id: int
    from_callsign: str
    to_callsign: str
    # Stored as string to be compatible with JSON, must be coerced to datetime.
    created_on: str
    subject: str
    body: str
    # Flag for indicating if the intended recipient has read the message.
    read: bool
    private: bool


@dataclass
class ReaderState:
    """
    State for tracking a connected reading user.
    """

    mode: Literal["home", "subject", "body"]
    to_callsign: str
    subject: str
    body: str
    private: bool


BoardConnections = dict[Connection, ReaderState]


class BoardState(TypedDict):
    """
    State for the bulletin board application.
    """

    # Key must be a string for JSON load and serialize compatibility,
    # but the key will be a stringified integer.
    version: Version
    messages: dict[str, Message]


class BoardSettings(TypedDict, total=False):
    """
    Settings for the bulletin board application.
    """

    welcome_message: str
    save_file_path: str
    max_message_length: int | None
    slots: int | None
    debug: bool



"""
Beacon service.
"""

import asyncio
import logging
from asyncio import Task
from collections.abc import Awaitable, Callable
from dataclasses import dataclass, field
from datetime import UTC, datetime, timedelta
from typing import TYPE_CHECKING

from pax25.ax25.address import Address, AddressHeader, Route
from pax25.ax25.constants import AX25_PID_TEXT, UFrameType
from pax25.ax25.control import Unnumbered
from pax25.ax25.frame import Frame
from pax25.types import BeaconServiceSettings
from pax25.utils import cancel_all, smart_clone

if TYPE_CHECKING:  # pragma: no cover
    from pax25.station import Station


logger = logging.getLogger(__name__)


type PerInterfaceBytes = dict[str, bytes]

type BeaconContents = bytes | PerInterfaceBytes | None

type BeaconCoroutine = Callable[["BeaconContext"], Awaitable[BeaconContents]]

type DigipeaterDict = dict[str, tuple[Address, ...]]


@dataclass(kw_only=True)
class BeaconSettings:
    """
    Data structure for defining a beacon. This structure is mutable and updating it will
    update the beacon's settings.
    """

    # Interval in seconds. Can be subsecond if needed via float.
    interval: int | float
    # Destination of resulting UI frame.
    dest: Address
    protocol: int = AX25_PID_TEXT
    max_length: int = 256
    update_station_timestamp: bool = False
    digipeaters: DigipeaterDict = field(default_factory=dict)
    coroutine: BeaconCoroutine


@dataclass(kw_only=True, frozen=True)
class BeaconContext:
    """
    Context for a beacon.
    """

    # The last time the station remembers running this beacon, or None if never.
    last_run: None | datetime = None
    # When the station first set up the beacon. The beacon may not have run at this
    # time.
    started: datetime
    # The registered name of this beacon.
    label: str
    # The settings for the beacon.
    spec: BeaconSettings
    station: "Station"


@dataclass(kw_only=True)
class BeaconTracker:
    """
    Tracking object for beacons-- notes when their last run was, and holds their spec.
    """

    spec: BeaconSettings
    started: datetime = field(default_factory=lambda: datetime.now(UTC))
    last_run: None | datetime = None


STATION_ID_LABEL = "station_id"
STATION_ID_CHECK_INTERVAL = 60


def delta_to_microseconds(delta: timedelta) -> int:
    """
    Converts a timedelta to microseconds.
    """
    return (
        ((delta.days * 24 * 60 * 60) + delta.seconds) * 1_000_000
    ) + delta.microseconds


async def next_interval(last_timestamp: datetime, interval: int | float) -> datetime:
    """
    Return the value of the next minute, after waiting for it to arrive.
    """
    next_interval_timestamp = last_timestamp + timedelta(seconds=interval)
    delta = next_interval_timestamp - last_timestamp
    delta_microseconds = delta_to_microseconds(delta)
    if delta_microseconds < 0:  # pragma: no cover
        # Must be right on the border between minutes. Unlikely, but possible.
        return next_interval_timestamp
    await asyncio.sleep(delta_microseconds / 1_000_000)
    return next_interval_timestamp


def generate_beacon_map(context: BeaconContext) -> PerInterfaceBytes:
    """
    Generate the default beacon text.
    """
    results: dict[str, bytes] = {}
    for interface in context.station.interfaces.values():
        text = context.station.name
        if not interface.gateway:
            continue
        apps = context.station.connection.application_map.get(interface.name, {})
        for key, value in apps.items():
            if not value.short_name:
                continue
            text += f" {key}/{value.short_name}"
        results[interface.name] = text.encode("utf-8")
    return results


async def station_id_beacon(context: BeaconContext) -> BeaconContents:
    """
    Example beacon which sends its text if we've been transmitting for a while
    and haven't sent a station ID.
    """
    interval = context.station.beacon._settings["id_beacon_interval"]
    if not context.station.frame_router.last_transmission:
        return None
    last_transmission = context.station.frame_router.last_transmission
    now = datetime.now(UTC)
    last_run = context.last_run or context.started
    if last_transmission > last_run and (last_run + timedelta(seconds=interval) < now):
        if context.station.beacon._settings["id_beacon_content"] is None:
            return generate_beacon_map(context)
        return context.station.beacon._settings["id_beacon_content"].encode("utf-8")
    return None


class BeaconService:
    """
    Beacon service. Used for sending out beacons on regular intervals.

    Beacons are sent out over all gateway interfaces.
    """

    def __init__(
        self,
        *,
        station: "Station",
        settings: BeaconServiceSettings | None = None,
    ):
        """
        Initialize the beacon service.
        """
        self._beacons: dict[str, BeaconTracker] = {}
        self.station = station
        self._settings: BeaconServiceSettings = BeaconServiceSettings(
            id_beacon_enabled=True,
            id_beacon_interval=600,
            id_beacon_destination="ID",
            id_beacon_digipeaters={},
            id_beacon_content=None,
        )
        if settings:
            self._settings.update(settings)
        self._id_beacon_tracker: None | BeaconTracker = None
        self._loop_tasks: dict[str, Task[None]] = {}
        self._tasks: set[Task[None]] = set()
        if self._settings["id_beacon_enabled"]:
            self._add_station_beacon()

    @property
    def settings(self) -> BeaconServiceSettings:
        """
        Return the beacon service's settings.
        """
        return smart_clone(self._settings)

    async def reload_settings(self, settings: BeaconServiceSettings) -> None:
        """
        Reload the beacon settings.
        """
        self._settings.update(settings)
        if self._settings["id_beacon_enabled"]:
            self._add_station_beacon()
        else:
            self._remove_station_beacon()
            return
        assert self._id_beacon_tracker
        self._id_beacon_tracker.spec = self._station_id_settings

    @property
    def _station_id_settings(self) -> BeaconSettings:
        return BeaconSettings(
            interval=STATION_ID_CHECK_INTERVAL,
            dest=Address.from_string(self._settings["id_beacon_destination"]),
            digipeaters={
                key: tuple(
                    Address.from_string(digipeater) for digipeater in digipeater_strings
                )
                for key, digipeater_strings in self._settings[
                    "id_beacon_digipeaters"
                ].items()
            },
            coroutine=station_id_beacon,
        )

    def _add_station_beacon(self) -> None:
        """
        Add the default station ID beacon.
        """
        if self._id_beacon_tracker:
            return
        self._id_beacon_tracker = self.add_beacon(
            STATION_ID_LABEL,
            self._station_id_settings,
        )

    def _remove_station_beacon(self) -> None:
        """
        Remove the default station ID beacon.
        """
        if not self._id_beacon_tracker:
            return
        self.remove_beacon(STATION_ID_LABEL)

    def run(self) -> None:
        """
        Initialization function. Starts the beacon service with the included station ID
        beacon.
        """
        for label, tracker in self._beacons.items():
            self._loop_tasks[label] = asyncio.ensure_future(
                self.beacon_loop(label, tracker)
            )

    def add_beacon(self, label: str, beacon_spec: BeaconSettings) -> BeaconTracker:
        """
        Adds a beacon for the service to track.
        """
        if label in self._beacons:
            logger.warning(
                f"Duplicate beacon {repr(label)} added. Overriding existing entry. "
                f"This is likely a bug."
            )
        tracker = BeaconTracker(
            spec=beacon_spec,
            last_run=None,
        )
        self._beacons[label] = tracker
        if self.station.running:
            self._loop_tasks[label] = asyncio.ensure_future(
                self.beacon_loop(label, tracker)
            )
        return tracker

    def remove_beacon(self, label: str) -> None:
        """
        Removes a beacon from the service.
        """
        if label not in self._beacons:
            logger.warning(
                f"Attempted to remove non-existent beacon {repr(label)}. "
                f"This is likely a bug."
            )
            return
        del self._beacons[label]
        if label in self._loop_tasks:
            # Might happen if called before the station has started.
            self._loop_tasks[label].cancel()
            del self._loop_tasks[label]

    async def resolve_beacon(
        self,
        label: str,
        future: Awaitable[BeaconContents],
    ) -> None:
        """
        Performs a run of the given beacon.
        """
        try:
            info = await future
        except Exception as err:
            logger.error(err, exc_info=True)
            return
        if info is None:
            return
        context = self._beacons[label]
        base_frame = Frame(
            route=Route(
                src=AddressHeader(address=Address(name=self.station.name)),
                dest=AddressHeader(address=context.spec.dest),
                digipeaters=tuple(),
            ),
            control=Unnumbered(frame_type=UFrameType.UNNUMBERED_INFORMATION),
            pid=context.spec.protocol,
        )
        length = context.spec.max_length - base_frame.size()
        if length < 0:
            length = 0
        to_send: PerInterfaceBytes = {}
        if isinstance(info, bytes):
            for interface in self.station.interfaces.values():
                if not interface.gateway:
                    continue
                to_send[interface.name] = info
        else:
            to_send = info
        for key, value in to_send.items():
            frame = base_frame._replace(
                info=value[:length],
                route=base_frame.route._replace(
                    digipeaters=tuple(
                        AddressHeader(address=digipeater)
                        for digipeater in context.spec.digipeaters.get(key, tuple())
                    ),
                ),
            )
            self.station.frame_router.send_frame(
                self.station.interfaces[key],
                frame,
                update_timestamp=context.spec.update_station_timestamp,
            )
        context.last_run = datetime.now(UTC)

    def clean_tasks(self) -> None:
        """
        Clear out any beacon promises that have already been fulfilled.
        """
        self._tasks = set(task for task in self._tasks if not task.done())

    async def beacon_loop(self, label: str, beacon: BeaconTracker) -> None:
        """
        Creates a loop for a beacon.

        This loop is configured to autocorrect for drift over time. Python won't give
        us true microsecond precision, but we can get close enough for any needs
        packet radio has. It might fail to correct if the interval is higher than we
        can reliably schedule recurring events, in which case this will just go as
        fast as it can.
        """
        # This loop is configured to autocorrect for drift over time. Python won't give
        # us true microsecond precision, but we can get close enough for any needs
        # packet radio has. It might fail to correct if the interval is higher than we
        # can reliably schedule recurring events, in which case this will just go as
        # fast as it can.
        next_timestamp = datetime.now(UTC)
        while True:
            next_timestamp = await next_interval(next_timestamp, beacon.spec.interval)
            self.clean_tasks()
            await self.resolve_beacon(
                label,
                beacon.spec.coroutine(
                    BeaconContext(
                        spec=beacon.spec,
                        station=self.station,
                        last_run=beacon.last_run,
                        started=beacon.started,
                        label=label,
                    )
                ),
            )

    async def shutdown(self) -> None:
        """
        Shutdown the beacons.
        """
        await cancel_all(self._loop_tasks.values())
        self._loop_tasks = {}
        self._beacons = {}



"""
Address-related data structures.
"""

from collections.abc import Iterator
from typing import NamedTuple, Self, cast

from pax25.ax25.constants import (
    AX25_ADDR_SIZE,
    AX25_CODEPAGE,
    AX25_ENDIAN,
    AX25_REPEATER_MAX,
    AX25_SSID_MASK_COMMAND,
    AX25_SSID_MASK_LAST_ADDRESS,
    AX25_SSID_MASK_RESERVED,
    AX25_SSID_MASK_SSID,
    AX25_SSID_SHIFT_RESERVED,
    AX25_SSID_SHIFT_SSID,
)
from pax25.ax25.exceptions import DisassemblyError
from pax25.protocols import JSONObj


class Address(NamedTuple):
    """
    Address information used for connections.
    """

    name: str
    ssid: int = 0

    def size(self) -> int:
        """
        Length, in bytes, of this structure when assembled.
        """
        return AX25_ADDR_SIZE

    @classmethod
    def disassemble(cls, data: bytes) -> Self:
        """
        Instantiate an Address from bytes. Will fail if the length of
        bytes is not precisely correct. Ignores flags on the SSID byte.
        """

        if len(data) != AX25_ADDR_SIZE:
            raise DisassemblyError(
                f"Address of incorrect length. Expected {AX25_ADDR_SIZE}, "
                f"got {len(data)} ({data!r})"
            )
        ssid = int((data[-1] & AX25_SSID_MASK_SSID) >> AX25_SSID_SHIFT_SSID)
        array = bytearray(data)
        # Addresses may only be valid ASCII values for upper-case letters or numbers.
        for i in range(0, AX25_ADDR_SIZE - 1):
            array[i] >>= 1
            if not (array[i] == 32 or (48 <= array[i] <= 57) or (65 <= array[i] <= 90)):
                raise DisassemblyError(f"ERROR: Corrupt address field: {array}")
        # ...But we decode as utf-8 so we don't end up with type errors later.
        name = array[: AX25_ADDR_SIZE - 1].decode("utf-8").strip()
        return cls(name=name, ssid=ssid)

    def assemble(self) -> bytes:
        """
        Takes the address and serializes it to a byte array.
        """
        data = bytearray()
        name = self.name.ljust(AX25_ADDR_SIZE - 1)
        data += bytes(name, AX25_CODEPAGE)
        for i in range(0, AX25_ADDR_SIZE - 1):
            data[i] <<= 1
        ssid_byte = 0
        ssid_byte = ssid_byte | (self.ssid << AX25_SSID_SHIFT_SSID)
        data += ssid_byte.to_bytes(1, AX25_ENDIAN)
        return bytes(data)

    @classmethod
    def from_pattern_string(cls, data: str) -> Iterator[Self]:
        """
        Generates an iterable of all matching strings for an address pattern.

        If a string is a normal address, like KW6FOX or K1LEO-2, it will return only
        that entry. However, if it's a pattern like KW6FOX-*, it will return all valid
        SSIDs for KW6FOX.

        In the _connect_future, we may support more patterns than
        """
        segments = data.split("-", maxsplit=1)
        if len(segments) < 2:
            yield cls.from_string(data)
            return
        if segments[1] == "*":
            for i in range(16):
                yield cls(name=segments[0], ssid=i)
            return
        yield cls.from_string(data)

    @classmethod
    def from_string(cls, data: str) -> Self:
        """
        Given a standard address string, create an Address. Address strings look like:
        KW6FOX
        K1LEO-3
        FOXBOX
        """
        if not data:
            raise ValueError("Empty strings are not valid addresses.")
        segments = data.split("-", 2)
        name = segments[0].upper()
        if 7 < len(name) > 0:
            raise ValueError(
                f"Names must be between one and six letters/numbers. "
                f"{repr(name)} is invalid."
            )
        if not name.isalnum():
            raise ValueError(f"Names must be alphanumeric. {repr(name)} is invalid.")
        try:
            name.encode("ascii")
        except ValueError as err:
            raise ValueError(
                f"Invalid character(s) found. Must be ASCII. {repr(name)} is invalid."
            ) from err
        ssid = 0
        if len(segments) == 2:
            try:
                ssid = int(segments[1])
            except ValueError as error:
                raise ValueError(
                    f"SSID must be an integer. Found: {repr(segments[1])}"
                ) from error
            if not 0 <= ssid <= 15:
                raise ValueError(
                    f"SSID must be between 0 and 15 inclusive. Found: {ssid}"
                )
        return cls(name=name, ssid=ssid)

    def __str__(self) -> str:
        """
        Form to string representation.
        """
        if self.ssid == 0:
            return self.name
        return f"{self.name}-{self.ssid}"

    def to_json(self) -> JSONObj:
        return {
            "__class__": self.__class__.__name__,
            "name": self.name,
            "ssid": self.ssid,
        }

    @classmethod
    def from_json(cls, obj: JSONObj) -> Self:
        kwargs = {
            "name": cast(str, obj["name"]),
            "ssid": cast(int, obj["ssid"]),
        }
        return cls(**kwargs)  # type: ignore[arg-type]


class AddressHeader(NamedTuple):
    """
    Address with metadata.
    """

    address: Address
    # These reserved bits are collected, but we don't have any particular use for them.
    # They can be used by a network for whatever purpose it deems useful.
    reserved: int = 3
    command_or_repeated: bool = False

    def size(self) -> int:
        """
        Length, in bytes, of this structure when assembled.
        """
        return AX25_ADDR_SIZE

    def __str__(self) -> str:
        string = str(self.address)
        if self.command_or_repeated:
            string += "*"
        if self.reserved != 3:
            string += f"(R{self.reserved})"
        return string

    @classmethod
    def disassemble(cls, data: bytes) -> Self:
        """
        Instantiate an AddressHeader from bytes. Will fail if the length of
        bytes is not precisely correct.
        """
        address = Address.disassemble(data)
        flags = data[-1]
        reserved = (flags & AX25_SSID_MASK_RESERVED) >> AX25_SSID_SHIFT_RESERVED
        command = bool(flags & AX25_SSID_MASK_COMMAND)
        return cls(
            address=address,
            reserved=reserved,
            command_or_repeated=command,
        )

    def assemble(self) -> bytes:
        """
        Assemble the AddressHeader into bytes suitable for transmission.
        """
        data = self.address.assemble()
        # Add bitmask to set the ch flag as necessary on the last byte.
        ssid_byte = data[-1]
        data = data[:-1]
        ssid_byte |= self.reserved << AX25_SSID_SHIFT_RESERVED
        ssid_byte |= self.command_or_repeated and AX25_SSID_MASK_COMMAND
        data += ssid_byte.to_bytes(1, AX25_ENDIAN)
        return data

    def to_json(self) -> JSONObj:
        return {
            "__class__": self.__class__.__name__,
            "address": self.address.to_json(),
            "reserved": self.reserved,
            "command_or_repeated": self.command_or_repeated,
        }

    @classmethod
    def from_json(cls, obj: JSONObj) -> Self:
        kwargs = {
            "address": Address.from_json(cast(JSONObj, obj["address"])),
            "reserved": cast(int, obj["reserved"]),
            "command_or_repeated": cast(bool, obj["command_or_repeated"]),
        }
        return cls(**kwargs)  # type: ignore[arg-type]


def is_last_address(byte: int) -> int:
    """
    Get the value of the ext flag from the address flags mask.
    """
    return bool(byte & AX25_SSID_MASK_LAST_ADDRESS)


class Route(NamedTuple):
    """
    Specification for a packet's intended route of traffic-- its source,
    its destination, and what digipeaters lay between.
    """

    src: AddressHeader
    dest: AddressHeader
    digipeaters: tuple[AddressHeader, ...] = tuple()

    def size(self) -> int:
        """
        Length, in bytes, of this structure when assembled.
        """
        return (2 + len(self.digipeaters)) * AX25_ADDR_SIZE

    def __str__(self) -> str:
        """
        Represents a route path for a frame in string form. Will add a * to the last
        repeater which has repeated the frame.
        """
        digi_section = ""
        if self.digipeaters:
            digi_section = ","
            digi_strings = []
            last_repeated_index = None
            for index, digi in enumerate(self.digipeaters):
                digi_strings.append(str(digi.address))
                if digi.command_or_repeated:
                    last_repeated_index = index
            if last_repeated_index is not None:
                digi_strings[last_repeated_index] += "*"
            digi_section += ",".join(digi_strings) + "/V"
        return "".join(
            [
                str(self.src.address),
                ">",
                str(self.dest.address),
                digi_section,
            ]
        )

    @classmethod
    def disassemble(cls, data: bytes) -> Self:
        """
        Instantiate a route path from a data byte array. If there is leftover data,
        it will be ignored.
        """
        dest_bytes = data[:AX25_ADDR_SIZE]
        dest = AddressHeader.disassemble(dest_bytes)
        data = data[AX25_ADDR_SIZE:]
        src_bytes = data[:AX25_ADDR_SIZE]
        src = AddressHeader.disassemble(src_bytes)
        last_address = is_last_address(src_bytes[-1])
        data = data[AX25_ADDR_SIZE:]
        digipeaters = []
        while len(data) and not last_address:
            digipeater_bytes = data[:AX25_ADDR_SIZE]
            digipeater = AddressHeader.disassemble(digipeater_bytes)
            last_address = is_last_address(digipeater_bytes[-1])
            digipeaters.append(digipeater)
            data = data[AX25_ADDR_SIZE:]
        if len(digipeaters) > AX25_REPEATER_MAX:
            raise DisassemblyError(
                f"Too many digipeaters specified. Maximum is 8. Received: {digipeaters}"
            )
        return cls(src=src, dest=dest, digipeaters=tuple(digipeaters))

    def assemble(self) -> bytes:
        """
        Assemble the route path into a byte array.
        """
        data = bytearray()
        to_join = (self.dest, self.src) + self.digipeaters
        for item in to_join:
            data.extend(item.assemble())
        # The 'last_address' flag must be set on the last address in the set to indicate
        # no further addresses are forthcoming.
        ssid_byte = int(data.pop())
        ssid_byte |= 1
        return bytes(data) + ssid_byte.to_bytes(1, AX25_ENDIAN)

    def to_json(self) -> JSONObj:
        return {
            "__class__": self.__class__.__name__,
            "src": self.src.to_json(),
            "dest": self.dest.to_json(),
            "digipeaters": [digipeater.to_json() for digipeater in self.digipeaters],
        }

    @classmethod
    def from_json(cls, obj: JSONObj) -> Self:
        kwargs = {
            "src": AddressHeader.from_json(cast(JSONObj, obj["src"])),
            "dest": AddressHeader.from_json(cast(JSONObj, obj["dest"])),
            "digipeaters": tuple(
                AddressHeader.from_json(digipeater)
                for digipeater in cast(list[JSONObj], obj["digipeaters"])
            ),
        }
        return cls(**kwargs)  # type: ignore[arg-type]



"""
Included generic parser functions which can be used by client developer commands.
"""

from collections.abc import Callable
from pathlib import Path
from typing import TYPE_CHECKING, TypeVar

from pax25.applications.autocomplete import AutocompleteDict
from pax25.ax25.address import Address

if TYPE_CHECKING:
    from pax25.applications.router import ParserSpec


def string_parser(spec: "ParserSpec") -> str:
    """
    Returns the stripped argument string without further modification.
    """
    return spec.args.strip()


def raw_string_parser(spec: "ParserSpec") -> str:
    """
    Does absolutely no processing to the arguments after the command name.

    NOTE: If you need the full input from the user, including the command, use
    the `raw_input` property on the CommandContext.
    """
    return spec.raw_input[len(spec.command) :]


def address_parser(spec: "ParserSpec") -> Address:
    """
    Parses an address from a string.
    """
    value = spec.args.strip().upper()
    if not value:
        raise ParseError("You must specify a callsign/address.")
    try:
        return Address.from_string(value)
    except ValueError as err:
        raise ParseError(f"{repr(value)} is not a valid callsign or address.") from err


def call_sign_parser(spec: "ParserSpec") -> Address:
    """
    Parses a callsign from a string, returning it as an Address with SSID 0.
    """
    raw_string = raw_string_parser(spec)
    if "-" in raw_string:
        raise ParseError("Do not specify SSID. Name only.")
    try:
        value = address_parser(spec)
    except (ValueError, ParseError) as err:
        raise ParseError("You must specify a valid callsign.") from err
    return value


def integer_parser(spec: "ParserSpec") -> int:
    """
    Parses an integer.
    """
    value = spec.args.strip()
    if not value:
        raise ParseError("You must specify an integer.")
    try:
        return int(spec.args.strip())
    except ValueError as err:
        raise ParseError(f"{repr(value)} is not a valid integer.") from err


def no_arguments(spec: "ParserSpec") -> None:
    """
    Use when a command accepts no arguments.
    """
    if spec.args.strip():
        raise ParseError("This command takes no arguments.")
    return None


def pull_segment(args: str) -> tuple[str, str]:
    """
    Splits a string once and returns the first segment as well as the remainder.
    """
    result = args.split(maxsplit=1)
    segment = result.pop(0)
    remainder = result[0] if result else ""
    return segment, remainder


def optional[T](
    func: Callable[["ParserSpec"], T],
) -> Callable[["ParserSpec"], None | T]:
    """
    Makes the expected arguments optional. If the args are empty, returns None.
    """

    def wrapped(spec: "ParserSpec") -> None | T:
        if not spec.args.strip():
            return None
        return func(spec)

    return wrapped


def requires_full_name[T](
    func: Callable[["ParserSpec"], T],
) -> Callable[["ParserSpec"], T]:
    """
    Wraps a command parser to first check that the full name of the command was used.
    """

    def wrapped(spec: "ParserSpec") -> T:
        if not spec.command_spec.command.lower() == spec.command.lower():
            raise ParseError(
                f"{repr(spec.command_spec.command)} must be typed in full."
            )
        return func(spec)

    return wrapped


class ParseError(Exception):
    """
    Throw when there is a parsing error.
    """


E = TypeVar("E", bound=str)


def autocompleted_enum(
    options: tuple[E, ...], *, default: E | None = None
) -> Callable[["ParserSpec"], E]:
    """
    A parser which will return a normalized entry.
    """
    lookup: AutocompleteDict[E] = AutocompleteDict()
    for entry in options:
        lookup[entry.lower()] = entry

    def parse_enum(spec: "ParserSpec") -> E:
        """
        Return which entry in the options was used, if any.
        """
        value = spec.args.strip().lower()
        if not value:
            if default is not None:
                return default
            raise ParseError(f"Argument must be one of: {options}")
        try:
            results = lookup[value]
            possibilities = list(sorted(result for result in results))
            if len(results) > 1:
                raise ParseError(
                    f"Ambiguous argument. Could be: {', '.join(possibilities)}"
                )
            result = possibilities[0]
            return result
        except KeyError as err:
            raise ParseError(f"Argument must be one of: {', '.join(options)}") from err

    return parse_enum


def file_path_parser(spec: "ParserSpec") -> Path:
    """
    Parser for files. Fails if given a directory, or the parent directory doesn't
    exist, but doesn't validate whether you can actually write to the file, or if it
    exists.
    """
    path = Path(spec.args.strip())
    try:
        if path.is_dir():
            raise ParseError(f"{repr(str(path))} is a directory.")
        if not path.parent.exists():
            raise ParseError("Parent directory does not exist.")
        return path
    except OSError as err:
        raise ParseError(str(err)) from err



"""
Loopback interface. Used internally for allowing internal connections without any
gateways being involved.
"""

import asyncio
from asyncio import Queue, Task
from collections.abc import Callable
from typing import TYPE_CHECKING, Self

from pax25.ax25.frame import Frame
from pax25.interfaces import Interface
from pax25.interfaces.types import LoopbackSettings
from pax25.services.connection.connection import Connection
from pax25.utils import cancel_all

if TYPE_CHECKING:  # pragma: no cover
    from pax25.services.connection.connection import Connection
    from pax25.station import Station


async def read_loop(queue: Queue[Frame], handler: Callable[[Frame], None]) -> None:
    """
    Starts a read loop for a connection and its associated queue.
    """
    while frame := await queue.get():
        handler(frame)


class LoopbackInterface(Interface[LoopbackSettings]):
    type = "Loopback"

    def __init__(self, name: str, settings: LoopbackSettings, station: "Station"):
        """
        Just stash the args but don't do anything with them.
        """
        self.name = name
        self._settings = settings
        self.station = station
        self.first_party_queue: Queue[Frame] = Queue()
        self.second_party_queue: Queue[Frame] = Queue()
        self._first_party_loop: Task[None] | None = None
        self._second_party_loop: Task[None] | None = None
        self._internal_connections: tuple[Connection, Connection] | None = None

    @property
    def listening(self) -> bool:
        """
        Returns a bool indicating whether the interface is listening.
        """
        tests = [
            self._first_party_loop and not self._first_party_loop.done(),
            self._second_party_loop and not self._second_party_loop.done(),
        ]
        return any(tests)

    @property
    def sudo(self) -> bool:
        """
        Returns if this connection should be privileged.
        """
        return self._settings["connection"].is_admin

    @property
    def connection(self) -> Connection:
        """
        Returns the initiating connection for this loopback.
        """
        return self._settings["connection"]

    @classmethod
    def install(
        cls, station: "Station", *, name: str, settings: LoopbackSettings
    ) -> Self:
        """
        Install and copy over references to the existing apps for the user's current
        connection.
        """
        instance = super().install(station, name=name, settings=settings)
        station.connection.application_map[instance.name] = (
            # Note: We're sharing a reference here. Theoretically, we should be updated
            # when the other interface updates.
            station.connection.application_map.get(
                settings["connection"].interface.name,
                {},
            )
        )
        return instance

    async def reload_settings(self, settings: LoopbackSettings) -> None:
        """
        Reloads the interface with revised settings.
        """
        raise NotImplementedError("Changing loopback settings is insane.")

    def register_connections(
        self, connection1: Connection, connection2: Connection
    ) -> None:
        """
        Gives the interface a reference to the internal connections so that it can
        sanely close them out on shutdown.
        """
        self._internal_connections = (connection1, connection2)

    @property
    def gateway(self) -> bool:
        """
        Loopbacks are never gateways.
        """
        return False

    def send_frame(self, frame: Frame) -> None:
        """
        Dummy send frame function. Does nothing.
        """

    def initialize_first_party(
        self,
        handle_frame: Callable[[Frame], None],
    ) -> Callable[[Frame], None]:
        """
        Builds the first party sender function, and starts its loop.
        """
        self._first_party_loop = asyncio.ensure_future(
            read_loop(self.first_party_queue, handle_frame)
        )

        def send_frame(frame: Frame) -> None:
            """
            Puts a frame from the first party on the second party's queue.
            """
            self.second_party_queue.put_nowait(frame)
            # We are bypassing the matchers when adding to the queue, but we still want
            # frames to be matchable for any novel use cases.
            self.station.frame_router.process_frame(self, frame)

        return send_frame

    def initialize_second_party(
        self,
        handle_frame: Callable[[Frame], None],
    ) -> Callable[[Frame], None]:
        """
        Builds the second party sender function, and starts its loop.
        """
        self._second_party_loop = asyncio.ensure_future(
            read_loop(self.second_party_queue, handle_frame)
        )

        def send_frame(frame: Frame) -> None:
            """
            Puts a frame from the first party on the second party's queue.
            """
            self.first_party_queue.put_nowait(frame)

        return send_frame

    def start(self) -> None:
        """
        Starts the read loop.
        """
        self.first_party_queue = Queue()
        self.second_party_queue = Queue()

    async def shutdown(self) -> None:
        """
        Dummy shut down function.
        """
        if self._internal_connections:
            for connection in self._internal_connections:
                connection.shutdown()
        await cancel_all([self._first_party_loop, self._second_party_loop])
        self.first_party_queue.shutdown()
        self.second_party_queue.shutdown()



"""
Helper utilities for application building.
"""

from collections import defaultdict
from collections.abc import Iterable

from pax25.services.connection.connection import Connection


def send_message(
    connection: Connection, message: str, append_newline: bool = True
) -> None:
    """
    Send a message string to a particular connection.
    """
    if append_newline:
        message += "\r"
    connection.send_bytes(message.encode("utf-8"))


def build_columns(
    entries_list: Iterable[str],
    num_columns: int = 6,
    column_width: int = 10,
) -> list[str]:
    """
    Build lines of help topic columns. Set num_columns to the number of columns to sort
    the entries into.
    """
    topic_sets = []
    current_set = []
    for entry in entries_list:
        current_set.append(entry)
        if len(current_set) == num_columns:
            topic_sets.append(current_set)
            current_set = []
    if current_set:
        topic_sets.append(current_set)
    # This might truncate some commands, but since there's autocomplete, it's
    # unlikely to matter.
    lines = []
    for topic_set in topic_sets:
        string = ""
        for entry in topic_set:
            string += entry[:column_width].ljust(column_width)
        string = string.rstrip()
        lines.append(string)
    return lines


def build_table(
    entries_list: Iterable[Iterable[str]],
) -> list[str]:
    """
    Build a table with automatic column sizing. You must set your own separator when
    joining them afterward.
    """
    column_sizes: defaultdict[int, int] = defaultdict(lambda: 0)
    # First, we must measure the column widths.
    for line in entries_list:
        for index, item in enumerate(line):
            if len(item) > column_sizes[index]:
                column_sizes[index] = len(item)
    lines = []
    for line in entries_list:
        lines.append(
            " ".join(
                (item.ljust(column_sizes[index]) for index, item in enumerate(line))
            )
        )
    return lines



"""
The file interface. This is used mostly for testing-- we can send in predefined
commands and let the system process them this way.

It's also useful for running apps over stdin/stdout, which are the default input/output
files.
"""

import asyncio
import logging
import os
import signal
import threading
from asyncio import Queue, QueueShutDown, Task
from collections.abc import Callable, Coroutine
from contextlib import suppress
from sys import stdin, stdout
from types import FrameType
from typing import IO, TYPE_CHECKING, Literal, TextIO

from pax25.ax25.address import Address, AddressHeader, Route
from pax25.ax25.constants import AX25_PID_TEXT, SFrameType, UFrameType
from pax25.ax25.control import Info, Supervisory, Unnumbered
from pax25.ax25.frame import Frame
from pax25.ax25.utils import is_command, response
from pax25.interfaces.types import FileSettings, Interface
from pax25.services.connection.variables import RolloverCounter
from pax25.utils import async_wrap, cancel_all, normalize_line_endings

if TYPE_CHECKING:  # pragma: no cover
    from pax25.station import Station


logger = logging.getLogger(__name__)


def build_threaded_reader(
    handle: IO[bytes], chunk: int
) -> tuple[Callable[[], Coroutine[None, None, bytes]], Callable[[], None]]:
    """
    Creates a thread and makes a reader based on it.
    """
    queue = Queue[bytes]()

    def read_loop() -> None:
        while True:
            data = handle.read(chunk)
            if not data:
                queue.shutdown()
                return
            try:
                queue.put_nowait(data)
            except QueueShutDown:
                return

    thread = threading.Thread(target=read_loop)
    thread.start()

    async def read_next() -> bytes:
        try:
            return await queue.get()
        except QueueShutDown:
            return b""

    def close() -> None:
        queue.shutdown()
        handle.close()
        thread.join()

    return read_next, close


class FileInterface(Interface[FileSettings]):
    """
    A file interface that will read in on a file and send the resulting
    bytes to another file. By default, input is stdin, and output is stdout.

    You may want to set the sudo settings flag to True if this is intended to be the
    administrative connection. Privileges will only be elevated if this flag is true
    AND the 'source ssid' argument matches the station name, which it will by default.
    """

    type = "File"

    def __init__(self, name: str, settings: FileSettings, station: "Station"):
        self.name = name
        self._settings = settings
        self.station = station
        self._input: IO[bytes] | None = None
        self._output: IO[bytes] | None = None
        self._output_tty = False
        self._close_input = False
        self._close_output = False
        self._read_loop: Task[None] | None = None
        self._write_loop: Task[None] | None = None
        self._vs = RolloverCounter(name="Send State Variable")
        self._vr = RolloverCounter(name="Receive State Variable")
        self._send_queue: Queue[Frame] = Queue()
        self._closer: Callable[[], None] = lambda: None
        self._route: None | Route = None

    @property
    def listening(self) -> bool:
        """
        Returns a bool indicating whether the interface is listening.
        """
        if not self._read_loop:
            return False
        return not self._read_loop.done()

    @property
    def gateway(self) -> bool:
        """
        Files should not be gateways.
        """
        return False

    @property
    def sudo(self) -> bool:
        """
        Whether connections inbound on this interface should be considered privileged.
        """
        return self._settings.get("sudo", False)

    async def reload_settings(self, settings: FileSettings) -> None:
        """
        Reload the file interface's settings.
        """
        await self.shutdown()
        self._settings = settings
        self._route = None
        self.start()

    def handle_for(
        self,
        file_path: str | IO[bytes] | None,
        mode: Literal["r", "w"],
        *,
        default_interface: TextIO,
    ) -> IO[bytes]:
        """
        Gets the file handle for a specified file path, falling back to an interface
        if the path is falsy.
        """
        handle: IO[bytes] = default_interface.buffer
        if file_path:
            if isinstance(file_path, str):
                handle = open(  # noqa: SIM115
                    file_path,
                    f"{mode}b",
                )
            else:
                handle = file_path
        if handle.isatty():  # pragma: no cover
            # We need to massage output if we're displaying to a user's terminal,
            # since pax25's native line endings are JUST carriage returns. This
            # means that on modern systems, be they Windows or *NIX, the line will
            # just keep overwriting itself.
            self._output_tty = True

        return handle

    @property
    def route(
        self,
    ) -> Route:
        """
        Shorthand function for adding the first and second party to the frames we
        construct.

        This is run many, many times a second in test so we do a bit of caching.
        """
        if self._route is None:
            self._route = Route(
                src=AddressHeader(
                    address=Address(
                        name=self._settings.get("source_name", self.station.name),
                        ssid=self._settings.get("source_ssid", 0),
                    )
                ),
                dest=AddressHeader(
                    address=Address(
                        name=self._settings.get("destination_name", self.station.name),
                        ssid=self._settings.get("destination_ssid", 0),
                    )
                ),
            )
        return self._route

    def send_receive_ready(
        self,
        modifier: Callable[[Frame], Frame] = lambda x: x,
    ) -> None:
        """
        Send receive ready. Potentially important if we're not sending but just
        receiving for a while.
        """
        self.station.frame_router.process_frame(
            self,
            modifier(
                Frame(
                    route=self.route,
                    control=Supervisory(
                        frame_type=SFrameType.RECEIVE_READY,
                        receiving_sequence_number=self._vr.value,
                    ),
                    pid=None,
                )
            ),
        )

    def send_frame(self, frame: Frame) -> None:
        """
        Add a frame to the queue for processing.
        """
        self._send_queue.put_nowait(frame)

    async def write_loop(self) -> None:
        """
        Handle the frames that have been sent back to us for writing.
        """
        while frame := await self._send_queue.get():
            self.handle_received_frame(frame)

    def handle_received_frame(self, frame: Frame) -> None:
        """
        Handles a frame from the Frame Router. We only care about a handful of frame
        types.
        """
        match frame.control:
            case Info():
                if not self._output:  # pragma: no cover
                    raise RuntimeError(
                        "Received a frame, but we don't have an open output file!",
                    )
                info = frame.info
                if self._output_tty:  # pragma: no cover
                    # Normalize line endings, in the case someone send
                    info = normalize_line_endings(info)
                self._output.write(info)
                self._output.flush()
                self._vr.increment()
                # Always send receive ready since we have zero lag time and may not have
                # another frame available.
                self.send_receive_ready()
            case Unnumbered() as control:
                match control.frame_type:
                    case UFrameType.DISCONNECT:
                        self.station.frame_router.process_frame(
                            self,
                            Frame(
                                route=self.route,
                                control=Unnumbered(
                                    frame_type=UFrameType.UNNUMBERED_ACKNOWLEDGE
                                ),
                                pid=None,
                            ),
                        )
                        asyncio.ensure_future(self.conditional_shutdown(remote=True))
            case Supervisory() as control:
                if (control.frame_type == SFrameType.RECEIVE_READY) and is_command(
                    frame
                ):
                    self.send_receive_ready(modifier=response)

    async def conditional_shutdown(self, remote: bool = False) -> None:
        """
        If set to shut down the station automatically, go ahead and do so.
        """
        if self._settings.get("auto_shutdown", True) or remote:
            if self.station.closing:
                # Already shutting down.
                return
            await self.station.shutdown()

    def start(self) -> None:
        """
        Starts the read loop and schedules it.
        """
        self._input = self.handle_for(
            self._settings.get("input"),
            "r",
            default_interface=stdin,
        )
        self._output = self.handle_for(
            self._settings.get("output"),
            "w",
            default_interface=stdout,
        )
        assert self._input
        assert self._output
        if self._input.isatty():

            def signal_handler(_signal: int, _frame: FrameType | None) -> None:
                asyncio.ensure_future(self.station.shutdown())

            # Register keyboard interrupt handler.
            signal.signal(signal.SIGINT, signal_handler)
        # This will create the connection in the connection table.
        self._read_loop = asyncio.ensure_future(self.read_loop())
        self._write_loop = asyncio.ensure_future(self.write_loop())

    async def read_loop(self) -> None:
        """Reads input from the keyboard and sends it for routing."""
        assert self._input
        # If we're reading from stdin, there could be long delays between characters
        # (keystrokes) so we read in 1 byte at a time as a full message frame. The
        # reads block in their own thread to allow other tasks to run. This prevents
        # the application from freezing while waiting for keyboard input.
        read_amount = 256 if self._input.seekable() else 1
        self.station.frame_router.process_frame(
            self,
            Frame(
                route=self.route,
                control=Unnumbered(frame_type=UFrameType.SET_ASYNC_BALANCED_MODE),
                pid=None,
            ),
        )
        read_data, self._closer = build_threaded_reader(self._input, read_amount)
        while True:
            # Note: When reading from a TTY, we don't get ANY characters until the
            # user presses the return key. We will eventually need to handle cases like
            # pressing ctrl+c to exit connection mode but that hasn't been implemented
            # yet.
            message = await read_data()
            if self._input.isatty():  # pragma: no cover
                # If we're reading in from a terminal, we need to normalize input.
                # Need to check how this behaves on Windows.
                if os.name == "nt":
                    message = message.replace(b"\r", b"")
                message = message.replace(b"\n", b"\r")
                if not message:
                    continue
            if not message:
                if self._settings.get("auto_shutdown", True):
                    # The file has ended.
                    await asyncio.sleep(
                        self._settings.get("pre_shutdown_sleep", 1) / 1000
                    )
                    self.station.frame_router.process_frame(
                        self,
                        Frame(
                            route=self.route,
                            control=Unnumbered(frame_type=UFrameType.DISCONNECT),
                            pid=None,
                        ),
                    )
                    await self.conditional_shutdown()
                break
            frame = Frame(
                route=self.route,
                control=Info(
                    receiving_sequence_number=self._vr.value,
                    sending_sequence_number=self._vs.value,
                ),
                pid=AX25_PID_TEXT,
                info=message,
            )
            self.station.frame_router.process_frame(
                self,
                frame,
            )
            self._vs.increment()

    async def shutdown(self) -> None:
        """Closes the read loop, closes open file handles."""
        self._closer()
        self._read_loop, self._write_loop = await cancel_all(
            [self._read_loop, self._write_loop]
        )
        if self._output:
            # Can happen if the file closes between the await and now.
            with suppress(ValueError):
                await async_wrap(self._output.flush)()
            self._output.close()



"""Station class. Defines the station parameters."""

import asyncio
import json
import logging
from asyncio import Future, Task
from typing import Any, Literal

from pax25.ax25.address import Address
from pax25.exceptions import ConfigurationError
from pax25.frame_router import FrameRouter
from pax25.interfaces import Interface
from pax25.services.beacon import BeaconService
from pax25.services.connection.service import ConnectionService
from pax25.services.digipeater import Digipeater
from pax25.services.monitor import Monitor
from pax25.types import StationConfig
from pax25.utils import cancel, cancel_all, first, maybe_update, smart_clone

logger = logging.getLogger(__name__)


class Station:
    """
    The main station class for Pax25. This class is intended to manage a physical
    station's digital affairs.
    """

    def __init__(
        self,
        *,
        config: StationConfig | None = None,
        config_file_path: str = "",
    ):
        """
        Initializes a station. Specifying a config dictionary configures the station
        with the config dictionary's values.

        Specifying a config_file_path will not load the config_file_path's values UNLESS
        the config parameter is set to None.

        The config_file_path is where configuration should be saved by administrative
        utilities (such as the contributed save command in the CommandLine app.) Can be
        left as an empty string to specify no intended save file.
        """
        # Timers may get cancelled and the station may shut down before they would
        # be cleaned up normally. We track these tasks so if we shut down while any
        # are still active, we can reap them manually.
        self.config_file_path = config_file_path
        if (config is None) and self.config_file_path:
            with open(config_file_path) as config_file:
                config = json.load(config_file)
        if config is None:
            raise ConfigurationError(
                "No configuration supplied. Please either set the config argument "
                "or provide a config_file_path."
            )
        self._active_future: Future[None] | None = None
        address = Address.from_string(config["name"])
        assert str(address) == config["name"], (
            "Station names must be all caps and have no SSID suffix."
        )
        self._settings = config
        self.interfaces: dict[str, Interface[Any]] = {}
        self.frame_router = FrameRouter(station=self)
        self.digipeater = Digipeater(station=self, settings=config.get("digipeater"))
        self.connection = ConnectionService(
            station=self,
            settings=config.get("connection"),
        )
        self.monitor = Monitor(station=self, settings=config.get("monitor"))
        self.beacon = BeaconService(station=self, settings=config.get("beacon"))
        self.clear_subsettings()
        self._cleanup_lock: Future[None] | None = None
        self._cleanup_task: Future[None] | None = None
        self._to_clean: list[Task[None]] = []
        self.closing = False

    @property
    def running(self) -> bool:
        """
        Returns if the station is currently running.
        """
        if self._active_future is None:
            return False
        return not self._active_future.done()

    @property
    def settings(self) -> StationConfig:
        """
        Returns the current configuration of the station.
        """
        settings = smart_clone(self._settings)
        # These are all smart cloned.
        settings["monitor"] = self.monitor.settings
        settings["beacon"] = self.beacon.settings
        settings["digipeater"] = self.digipeater.settings
        settings["connection"] = self.connection.settings
        return settings

    @property
    def name(self) -> str:
        return self._settings["name"].upper()

    def clear_subsettings(self) -> None:
        """
        Settings are set on the station object, but the authoritative location of the
        settings for each service are on that service. Remove these keys so that if we
        erroneously rely on the internal representation of the settings for a service,
        we raise instead of using data which may be out of date, which could cause a
        very subtle bug.
        """
        keys: tuple[Literal["monitor", "beacon", "digipeater", "connection"], ...] = (
            "monitor",
            "beacon",
            "digipeater",
            "connection",
        )
        for key in keys:
            if key in self._settings:
                del self._settings[key]

    def collect_task(self, task: Task[None]) -> None:
        """
        Tasks which are 'fire and forget' can be sent here to be cleaned up later
        as needed. This prevents the garbage collector from freaking out and sending
        tracebacks as tasks from functions like those in the timer module get reaped.
        """
        self._to_clean.append(task)

    async def reload_settings(self, settings: StationConfig) -> None:
        """
        Reloads station configuration, and all components involved.
        """
        address = Address.from_string(settings["name"])
        assert str(address) == settings["name"], (
            "Station names must be all caps and have no SSID suffix."
        )
        self._settings = settings
        digipeater_settings = maybe_update(
            self.digipeater.settings,
            settings.get("digipeater"),
        )
        monitor_settings = maybe_update(
            self.monitor.settings,
            settings.get("monitor"),
        )
        beacon_settings = maybe_update(
            self.beacon.settings,
            settings.get("beacon"),
        )
        connection_settings = maybe_update(
            self.connection.settings,
            settings.get("connection"),
        )
        self.clear_subsettings()
        await asyncio.gather(
            self.digipeater.reload_settings(digipeater_settings),
            self.monitor.reload_settings(monitor_settings),
            self.beacon.reload_settings(beacon_settings),
            self.connection.reload_settings(connection_settings),
        )

    async def finished(self) -> None:
        """
        Await this to wait until the server is closed down programmatically.
        """
        if self._active_future is None:
            raise RuntimeError(
                "Station has not run. We cannot wait for it to finish if "
                "it never began!"
            )
        await self._active_future

    async def _cleanup_loop(self) -> None:
        """
        Cleanup loop.
        """
        sleep: Task[None] | None = None
        while self._cleanup_lock:
            sleep = Task(asyncio.sleep(0.500))
            await first(sleep, self._cleanup_lock)
            self._to_clean = [task for task in self._to_clean if not task.done()]
            await cancel(sleep)
        await cancel_all(self._to_clean)
        self._cleanup_task = None

    def run(self) -> Future[None]:
        """
        Starts the station. Tasks will not begin unless/until the asyncio loop is
        running.
        """
        if self._active_future and not self._active_future.done():
            raise RuntimeError("Station is already running!")
        self.monitor.run()
        self.bring_up_interfaces()
        self.beacon.run()
        self._cleanup_lock = Future()
        self._cleanup_task = asyncio.ensure_future(self._cleanup_loop())
        self._active_future = Future()
        return self._active_future

    async def start(self) -> None:
        """
        Starts the station and awaits its completion. Useful for projects that just
        have one station and want to use it as the event loop entry point, which is
        most projects, or anything which might need to wait on an entire station's
        lifecycle.
        """
        await self.run()

    def bring_up_interfaces(self) -> None:
        """
        Attempts to bring up all interfaces and queue them into the event loop.
        """
        for interface in self.interfaces.values():
            interface.start()

    def add_interface(self, name: str, interface: Interface) -> None:
        """
        Adds an interface to the station.
        """
        if interface.station != self:
            raise ConfigurationError(
                "You cannot assign an interface to an irrelevant station."
            )
        if not name:
            raise ConfigurationError("Interface name may not be blank.")
        if name.split() != [name]:
            raise ConfigurationError(
                f"Interface names may not have spaces. Got: {repr(name)}",
            )
        try:
            name.encode("ascii")
        except ValueError as err:
            raise ConfigurationError(
                f"Interface names MUST be ASCII. Got: {repr(name)}",
            ) from err
        if name in self.interfaces:
            raise ConfigurationError(f"Interface {repr(name)} already exists!")
        self.interfaces[name] = interface
        self.monitor.refresh_ports_cache()
        if self.running and not self.closing:
            interface.start()

    async def remove_interface(self, name: str) -> None:
        """
        Removes an interface from the station.
        """
        if name not in self.interfaces:
            raise KeyError(f"Interface {repr(name)} does not exist.")
        interface = self.interfaces[name]
        del self.interfaces[name]
        self.monitor.refresh_ports_cache()
        await asyncio.ensure_future(interface.shutdown())

    def get_nth_gateway(self, index: int) -> "Interface":
        """
        Get the nth gateway or raise an IndexError if not present.
        """
        if index <= 0:
            raise IndexError("Port numbers start at 1.")
        current = 0
        for interface in self.interfaces.values():
            if interface.gateway:
                current += 1
                if current == index:
                    return interface
        if not current:
            raise IndexError("No gateways available.")
        raise IndexError(f"Port {index} not found.")

    async def shutdown(self) -> None:
        """
        Shut down all interfaces, cutting off the station.
        """
        if not self.running:
            # Already shut down.
            return
        if self.closing and self._active_future:  # pragma: no cover
            logger.warning(
                "Shutdown called while shutdown already called. "
                "This may cause a deadlock.",
            )
            await self._active_future
            return
        self.closing = True
        await self.connection.shutdown()
        await self.beacon.shutdown()
        for interface in self.interfaces.values():
            await interface.shutdown()
        await self.monitor.shutdown()
        if self._active_future:
            self._active_future.set_result(None)
        if self._cleanup_task and self._cleanup_lock:
            self._cleanup_lock.cancel()
            self._cleanup_lock = None
            await self._cleanup_task
        self.closing = False



"""
Types for the command line app.
"""

from dataclasses import dataclass, field
from enum import Enum
from typing import TYPE_CHECKING, TypedDict

from pax25.interfaces.loopback import LoopbackInterface
from pax25.services.connection.connection import Connection
from pax25.utils import EnumReprMixin

if TYPE_CHECKING:  # pragma: no cover
    from pax25.contrib.command.command import CommandLine


class CommandSettings(TypedDict, total=False):
    """
    Settings for the command line application.
    """

    # Whether to automatically drop a user after they've disconnected from a remote.
    # Used mostly for testing, but may also work when acting as a node in the future.
    auto_quit: bool


class RemoteConnectionState(EnumReprMixin, Enum):
    """
    Enum for remote connection state.
    """

    FOREGROUND = 0
    BACKGROUND = 1


@dataclass()
class CommandLineState:
    """
    Keeps track of the current state for connected sessions.
    """

    remote_state: RemoteConnectionState = RemoteConnectionState.BACKGROUND
    remote_buffer: bytearray = field(default_factory=bytearray)
    connection: Connection | None = None
    # Used in loopbacks.
    dest_connection: Connection | None = None
    loopback_interface: LoopbackInterface | None = None


@dataclass()
class ShimSettings:
    """
    Settings for the shim application.
    """

    proxy_connection: Connection
    upstream_application: "CommandLine"



"""
Command line interface for controlling Pax25.
"""

import asyncio
from collections.abc import Callable
from contextlib import suppress
from dataclasses import dataclass
from importlib.metadata import version
from typing import cast
from uuid import uuid4

from pax25 import Application, Station
from pax25.applications.help import Help
from pax25.applications.parsers import (
    ParseError,
    no_arguments,
    pull_segment,
)
from pax25.applications.router import (
    CommandContext,
    CommandRouter,
    CommandSpec,
    ParserSpec,
)
from pax25.applications.utils import send_message
from pax25.ax25.address import Address
from pax25.ax25.constants import AX25_ENDIAN, AX25_REPEATER_MAX
from pax25.ax25.frame import Frame
from pax25.ax25.utils import roll_back_ssid
from pax25.contrib.command.commands.admin import (
    build_load_command,
    build_name_command,
    build_save_command,
    build_save_file_command,
    build_shutdown_command,
)
from pax25.contrib.command.commands.utilities import (
    build_ports_command,
    build_whoami_command,
)
from pax25.contrib.command.shim import ShimApplication
from pax25.contrib.command.types import (
    CommandLineState,
    CommandSettings,
    RemoteConnectionState,
    ShimSettings,
)
from pax25.interfaces import FileInterface, Interface
from pax25.interfaces.loopback import LoopbackInterface
from pax25.services.connection.connection import (
    Connection,
    ConnectionStatus,
    connection_key,
)
from pax25.types import Version
from pax25.utils import gateways_for

PAX25_VERSION = version("pax25")


def default_command_line_settings() -> CommandSettings:
    """
    Generate default settings for the command line application.
    """
    return {"auto_quit": False}


def digipeaters_from_string(string: str) -> tuple[Address, ...]:
    """
    Given a string like 'via SOME-1,THING-2,WAT-3', determine digipeaters.
    """
    string = string.strip()
    if not string:
        return tuple()
    try:
        via, *args = string.split(maxsplit=1)
        if not via.lower() == "via" or not args:
            raise ValueError
    except ValueError as err:
        raise ValueError(
            "Digipeater list broken. Try something like 'via NAME-1,NAME-2'."
        ) from err
    string = args[0]
    addresses = tuple(
        Address.from_string(raw_address.strip()) for raw_address in string.split(",")
    )
    if len(addresses) > AX25_REPEATER_MAX:
        raise ValueError(f"Too many digipeaters. Max is {AX25_REPEATER_MAX}.")
    return addresses


def get_indexed_gateway(station: Station, index: int) -> Interface:
    """
    Wrapper for get_nth_gateway that re-raises an IndexError as a parsing error.
    """
    try:
        return station.get_nth_gateway(index)
    except IndexError as err:
        raise ParseError(str(err)) from err


@dataclass(kw_only=True, frozen=True)
class MakeConnectionSpec:
    """
    Used by the connection parser to pass around distilled arguments.
    """

    destination: Address
    digipeaters: tuple[Address, ...]
    interface: Interface
    loopback: bool


def create_connection_parser(
    station: Station,
) -> Callable[[ParserSpec], MakeConnectionSpec]:
    """
    Generate a parser for the connection commands. Has to be generated since the
    available interfaces/ports are station-specific.
    """

    def connection_parser(spec: ParserSpec) -> MakeConnectionSpec:
        raw_args = spec.args.strip()
        if not raw_args:
            raise ParseError("Need station address.")

        explicit_interface = False
        proposed_interface: Interface | None = None
        interface: Interface | None = None
        loopback = False
        first_arg, remainder = pull_segment(raw_args)
        # First arg can be either a port number/alias, or the destination station.
        # If no port/alias is specified, then it grabs the default gateway (the first
        # registered interface where gateway is set to True)
        try:
            port_number = int(first_arg)
            proposed_interface = get_indexed_gateway(station, port_number)
            raw_destination, remainder = pull_segment(remainder)
            explicit_interface = True
        except ValueError:
            if first_arg in station.interfaces:
                proposed_interface = station.interfaces[first_arg]
                raw_destination, remainder = pull_segment(remainder)
            else:
                with suppress(ParseError):
                    proposed_interface = get_indexed_gateway(station, 1)
                raw_destination = first_arg
        try:
            destination = Address.from_string(raw_destination)
        except ValueError as err:
            raise ParseError(str(err)) from err
        if remainder:
            try:
                digipeaters = digipeaters_from_string(remainder)
            except ValueError as err:
                raise ParseError(str(err)) from err
        else:
            digipeaters = tuple()
        if not explicit_interface:
            try:
                application = station.connection.app_for(
                    spec.connection.interface,
                    destination,
                )
                if application == spec.connection.application:
                    # Don't allow connecting back to the current application to
                    # avoid infinite internal traversal.
                    raise ParseError(
                        "Cowardly refusing to connect to the current application "
                        "from the current application."
                    )
                interface = spec.connection.interface
                loopback = True
            except KeyError:
                pass
        if not interface:
            if proposed_interface is not None:
                interface = proposed_interface
            else:
                raise ParseError(
                    "No gateways available, or unrecognized internal address."
                )
        return MakeConnectionSpec(
            destination=destination,
            digipeaters=digipeaters,
            interface=interface,
            loopback=loopback,
        )

    return connection_parser


class CommandLine(Application[CommandSettings]):
    """
    Command line application. Used to handle things like connecting to other nodes
    or changing settings at runtime.
    """

    connections: dict[Connection, CommandLineState]
    frame_log: list[Frame]
    version = Version(major=0, minor=4, patch=0)
    version_string = (
        f"[PAX25-CLI-{version['major']}.{version['minor']}.{version['patch']}]"
    )
    # Used in the station ID beacon. The command line app also acts as the node app.
    short_name = "N"
    auto_quit: bool
    router: CommandRouter

    def setup(self) -> None:
        """
        Initial setup.
        """
        self.connections = {}
        settings = default_command_line_settings()
        settings.update(self.settings or {})
        self.settings = settings
        self.auto_quit = settings["auto_quit"]
        # Need to match all frames when monitoring.
        self.station.monitor.add_listener("cmd_monitor", self.frame_listener)
        self.router = CommandRouter(post_command_func=self.send_prompt)
        self.admin_router = CommandRouter(post_command_func=self.send_prompt)
        self.router.add(Help(self.router).spec)
        self.admin_router.add(Help(self.admin_router).spec)
        commands = (
            CommandSpec(
                command="apps",
                help="\r".join(
                    [
                        "Application lister. Shows all available applications you can "
                        "connect to on the local station, with addresses and info.",
                    ]
                ),
                parser=no_arguments,
                function=self.list_applications,
            ),
            CommandSpec(
                command="connect",
                help="\r".join(
                    [
                        "Use connect to connect to an outbound station. Examples:",
                        "c FOXBOX            # Connect to a station named FOXBOX, by "
                        "default on SSID 0",
                        "c KW6FOX-3          # Connect to SSID 3 on a station named "
                        "KW6FOX",
                        "c KW6FOX-2 via BOOP # Connect to KW6FOX-2 through an "
                        "intermediary station named BOOP",
                        "c 2 KW6FOX-2 via BOOP # Use port 2 to connect to KW6FOX "
                        "through an intermediary station named BOOP",
                        "",
                        "If no port is specified, checks local application "
                        "addresses. If not an internal address, uses port 1.",
                        "See also: PORTS",
                    ]
                ),
                parser=create_connection_parser(self.station),
                function=self.connect,
            ),
            CommandSpec(
                command="discon",
                # Full name is too big for command listing, so making it an alias.
                aliases=("disconnect",),
                help="Drops remote connection, if there is one.",
                parser=no_arguments,
                function=self.disconnect_remote,
            ),
            build_ports_command(self.station),
            CommandSpec(
                command="foreground",
                help="Returns you to the remote connected session, "
                "if one has been backgrounded.",
                parser=no_arguments,
                function=self.foreground_connection,
                aliases=("fg", "k"),
            ),
            build_whoami_command(),
            CommandSpec(
                command="quit",
                help="Closes the session.",
                aliases=("bye",),
                parser=no_arguments,
                function=self.quit,
            ),
            self.station.monitor.command,
        )
        self.router.add(*commands)
        self.admin_router.add(*commands)
        self.admin_router.add(
            build_shutdown_command(self.station),
            build_save_command(self.station),
            build_load_command(self.station),
            build_save_file_command(self.station),
            build_name_command(self.station),
        )

    def prompt_text(self, connection: Connection) -> str:
        """
        Generates the prompt text for the current connection. We serve slightly
        different prompt texts depending on whether this is a superuser connection
        or not.
        """
        if connection.is_admin:
            return "cmd:"
        else:
            return "ENTER COMMAND: B,C,J or Help ?\r"

    def send_prompt(self, connection: Connection) -> None:
        """
        Send the command prompt.
        """
        state = self.connections.get(connection)
        if not state:
            return
        if state.remote_state == RemoteConnectionState.FOREGROUND:
            return
        send_message(connection, self.prompt_text(connection), append_newline=False)

    def frame_listener(self, frame: Frame, _interface: Interface) -> bool:
        """
        Frame listener. Subscribes to the monitoring service and consumes frames in
        real time if there's a suitable, connected user.
        """
        for key, value in self.connections.items():
            if key.is_admin and isinstance(key.interface, FileInterface):  # noqa: SIM102
                if not (
                    value.connection or self.connection_state_table[key]["command"]
                ):
                    # Send any outstanding.
                    self.dump_logs_to_connection(key)
                    send_message(key, str(frame))
                    return True
        return False

    def dump_logs_to_connection(self, connection: Connection) -> None:
        """
        Send every log to the connection specified. Used when we disconnect and want
        to see what happened in the interim.
        """
        if not connection.is_admin and isinstance(connection.interface, FileInterface):
            return
        logs = self.station.monitor.consume_logs()
        if not logs:
            return
        prefix = len(gateways_for(self.station)) > 1
        for logged_frame in logs:
            frame_string = f"{logged_frame.port.number}:" if prefix else ""
            frame_string += str(logged_frame.frame)
            send_message(connection, frame_string)

    def on_proxy_bytes(self, connection: Connection, message: bytes) -> None:
        """
        Called by the shim application to indicate that we need to send bytes downstream
        to the connecting station.
        """
        state = self.connections[connection]
        if state.remote_state == RemoteConnectionState.BACKGROUND:
            state.remote_buffer.extend(message)
            return
        connection.send_bytes(message)

    def list_applications(
        self,
        connection: Connection,
        _context: CommandContext[None],
    ) -> None:
        """
        Lists out all internal applications available to the user.
        """
        apps = self.station.connection.application_map.get(
            connection.interface.name, {}
        )
        results = [
            f"{key}: {value.name}" for key, value in apps.items() if value != self
        ]
        send_message(connection, "\r".join(results))

    def disconnect_remote(
        self,
        connection: Connection,
        _context: CommandContext[None],
    ) -> None:
        """
        Disconnects from the remote station manually.
        """
        state = self.connections[connection]
        if not state.connection:
            send_message(connection, "Not connected. Can't disconnect.")
            return
        state.connection.disconnect()

    def on_proxy_killed(
        self,
        connection: Connection,
        jump_connection: Connection,
    ) -> None:
        """
        Called by the shim application to indicate that the connection has been killed.
        """
        if connection in self.connections:
            state = self.connections[connection]
            state.connection = None
            if state.dest_connection:
                state.dest_connection.disconnect()
            if state.loopback_interface:
                asyncio.ensure_future(state.loopback_interface.shutdown())
            self.dump_connection_buffer(connection)
            state.remote_state = RemoteConnectionState.BACKGROUND
        send_message(
            connection, f"*** Disconnected from {jump_connection.second_party}"
        )
        self.dump_logs_to_connection(connection)
        if self.auto_quit:
            connection.disconnect()

    def set_up_connection(
        self,
        *,
        source_connection: Connection,
        destination: Address,
        interface: Interface,
        digipeaters: tuple[Address, ...],
        loopback: bool,
    ) -> None:
        """
        Set up an outbound connection to route the user through.
        """
        state = self.connections[source_connection]
        if state.connection:
            send_message(
                source_connection,
                "You already have an active outbound connection. Use the 'disconnect' "
                "command to close out that connection before starting a new one.",
            )
            return
        if source_connection.first_party == source_connection.second_party:
            # Internal connection. No need to cycle the SSID.
            source = source_connection.first_party
        else:
            source = roll_back_ssid(source_connection.first_party)
        key = connection_key(source, destination, interface)
        if key in self.station.connection.table:
            send_message(source_connection, "Route busy. Try again later.")
            return
        send_message(source_connection, f"Connecting to {destination}...")
        shim_application = ShimApplication(
            name=f"(Shim for {key})",
            station=self.station,
            settings=ShimSettings(
                proxy_connection=source_connection, upstream_application=self
            ),
        )
        dest_connection: Connection | None = None
        loopback_interface: LoopbackInterface | None = None
        if loopback:
            loopback_interface = LoopbackInterface.install(
                self.station,
                name=f"__loopback-{uuid4()}",
                settings={"connection": source_connection},
            )

            def dest_handle_frame(frame: Frame) -> None:
                assert loopback_interface
                local_connection = cast(Connection, dest_connection)
                local_connection.inbound_frame(frame, loopback_interface)

            def source_handle_frame(frame: Frame) -> None:
                assert loopback_interface
                connection.inbound_frame(frame, loopback_interface)

            source_send_frame = loopback_interface.initialize_first_party(
                source_handle_frame,
            )
            dest_send_frame = loopback_interface.initialize_second_party(
                dest_handle_frame,
            )
            application = self.station.connection.app_for(interface, destination)
            dest_connection = Connection(
                station=self.station,
                first_party=source,
                is_admin=source_connection.is_admin,
                second_party=destination,
                digipeaters=tuple(),
                interface=loopback_interface,
                inbound=True,
                application=application,
                frame_sender=dest_send_frame,
            )
            # Hack the connection status to be where we need it to be upon starting.
            dest_connection.status = ConnectionStatus.CONNECTED
            connection = Connection(
                station=self.station,
                first_party=source,
                second_party=destination,
                is_admin=source_connection.is_admin,
                digipeaters=tuple(),
                interface=loopback_interface,
                inbound=False,
                application=shim_application,
                frame_sender=source_send_frame,
            )
            loopback_interface.register_connections(connection, dest_connection)
        else:
            connection = self.station.connection.add_connection(
                first_party=source,
                second_party=destination,
                digipeaters=digipeaters,
                interface=interface,
                inbound=False,
                application=shim_application,
            )
        state.connection = connection
        state.dest_connection = dest_connection
        state.loopback_interface = loopback_interface
        state.remote_state = RemoteConnectionState.FOREGROUND
        connection.negotiate()

    def connect(
        self,
        connection: Connection,
        context: CommandContext[MakeConnectionSpec],
    ) -> None:
        """
        Connect to a remote station.
        """
        self.set_up_connection(
            source_connection=connection,
            interface=context.args.interface,
            destination=context.args.destination,
            digipeaters=context.args.digipeaters,
            loopback=context.args.loopback,
        )

    def quit(self, connection: Connection, _context: CommandContext[None]) -> None:
        """
        Closes the application.
        """
        send_message(connection, "Goodbye!")
        connection.disconnect()

    def router_for(self, connection: Connection) -> CommandRouter:
        """
        Gets the appropriate router for the current connection.
        """
        if connection.is_admin:
            return self.admin_router
        return self.router

    def run_home_command(self, connection: Connection, message: str) -> None:
        """
        Match a message for a home command.
        """
        self.router_for(connection).route(connection, message)

    def on_startup(self, connection: Connection) -> None:
        """
        Set up the current connection's state.
        """
        self.connections[connection] = CommandLineState()
        self.dump_logs_to_connection(connection)
        if connection.is_admin:
            banner = f"PAX25 v{PAX25_VERSION} CLI, AGPLv3"
        else:
            banner = self.version_string
        send_message(
            connection,
            f"{banner}\r{self.prompt_text(connection)}",
            append_newline=False,
        )

    def dump_connection_buffer(self, connection: Connection) -> None:
        """
        Dump any outstanding data from the connected buffer to the user.
        """
        state = self.connections[connection]
        data = bytes(state.remote_buffer)
        state.remote_buffer = bytearray()
        send_message(
            connection,
            data.decode("utf-8", errors="backslashreplace"),
            append_newline=False,
        )

    def foreground_connection(
        self,
        connection: Connection,
        _: CommandContext[None],
    ) -> None:
        """
        Foregrounds a backgrounded connection.
        """
        state = self.connections[connection]
        if not state.connection:
            send_message(connection, "Not connected.")
            return
        state.remote_state = RemoteConnectionState.FOREGROUND
        self.dump_connection_buffer(connection)

    def background_connection(self, connection: Connection) -> None:
        """
        Sends the current connection to the background, if it exists. This brings us
        back to the command line.
        """
        state = self.connections[connection]
        state.remote_state = RemoteConnectionState.BACKGROUND
        self.dump_logs_to_connection(connection)

    def on_bytes(self, connection: Connection, data: bytes) -> None:
        """
        Processes incoming bytes, allowing for an interrupt to pull us out of a proxy
        connection.
        """
        for byte in data:
            if byte == 0x0B:
                self.background_connection(connection)
                continue
            super().on_bytes(connection, byte.to_bytes(1, AX25_ENDIAN))

    def on_message(self, connection: Connection, message: str) -> None:
        """
        Handle command from the user.
        """
        state = self.connections[connection]
        if (jump_connection := state.connection) and (
            state.remote_state == RemoteConnectionState.FOREGROUND
        ):
            jump_connection.send_bytes((message + "\r").encode("utf-8"))
            return
        # Send any outstanding frames that entered while they were typing before
        # giving them a response, so we're not missing any.
        self.dump_logs_to_connection(connection)
        self.run_home_command(connection, message)

    def on_shutdown(self, connection: Connection) -> None:
        """
        Clean up connection state.
        """
        state = self.connections[connection]
        if state.connection:
            state.connection.disconnect()
        if state.dest_connection:
            state.dest_connection.disconnect()
        del self.connections[connection]



"""
Connections handle state and method for connections between two stations (or,
in some cases, one part of a station to another part of itself). They are
instantiated by the Frame Router, and then handed to applications for
manipulation.
"""

import logging
from collections.abc import Callable
from dataclasses import dataclass, field, fields
from enum import Enum
from typing import TYPE_CHECKING, Any, Literal, Optional, cast

from pax25.ax25.address import Address, AddressHeader, Route
from pax25.ax25.constants import AX25_PID_TEXT, AX25_SEQ_MAX, SFrameType, UFrameType
from pax25.ax25.control import Info, Supervisory, Unnumbered
from pax25.ax25.frame import Frame
from pax25.ax25.matchers import (
    Matcher,
    check_all,
    has_dest_address,
    has_src_address,
    has_these_digipeaters,
    repeats_completed,
)
from pax25.ax25.utils import (
    build_receive_modifier,
    command,
    is_command,
    response,
    response_frame,
)
from pax25.services.connection.variables import FrameStateTracker
from pax25.timer import Timer, delay_action, retry_loop
from pax25.types import ConnectionSettings
from pax25.utils import (
    EnumReprMixin,
    digipeater_factor,
    smart_clone,
)

if TYPE_CHECKING:  # pragma: no cover
    from pax25.applications import BaseApplication
    from pax25.interfaces import Interface
    from pax25.services.connection.service import (
        ConnectionService,
    )
    from pax25.station import Station


type ConnectionKey = tuple[int, int, int]


logger = logging.getLogger(__name__)


class ConnectionStatus(EnumReprMixin, Enum):
    """
    Connection status flags. The integers here match the constants direwolf uses,
    but I'm not sure that it matters so long as Python can tell them apart,
    since I don't imagine we need their values for anything in particular.
    """

    DISCONNECTED = 0
    AWAIT_CONNECTION = 1
    AWAIT_DISCONNECTION = 2
    CONNECTED = 3


class ConnectionTimeout(Exception):
    """
    Exception thrown when failing to negotiate connection.
    """


class Connection:
    """
    Connections handle state and configuration for a virtual circuit between
    stations.

    If sudo is True, signals to applications that this connection is safe to assume
    as a superuser if the connected name matches the station name.
    """

    def __init__(
        self,
        *,
        first_party: Address,
        second_party: Address,
        digipeaters: tuple[Address, ...],
        interface: "Interface[Any]",
        application: Optional["BaseApplication[Any]"],
        station: "Station",
        inbound: bool,
        is_admin: bool = False,
        service: Optional["ConnectionService"] = None,
        frame_sender: Callable[[Frame], None] | None = None,
        close_callback: Callable[["Connection"], None] = lambda x: None,
        settings: ConnectionSettings | None = None,
    ):
        # The party that initializes the connection is the first party.
        self.first_party = first_party
        # The party that receives the initial connection request is the second party.
        self.second_party = second_party
        # Only one interface per connection. If there is a bridge, it needs to be done
        # via two connections with some kind of middleware.
        self.interface = interface
        # The set of digipeaters between the first and second party. Reverse these when
        # inbound.
        self.digipeaters = digipeaters
        # Used when running the negotiation task. If we're inbound, we need to do the
        # response frame. Otherwise, we send the initial connection frame.
        self.inbound = inbound
        self.send_buffer: bytearray = bytearray()
        self.is_admin = is_admin
        self.application = application
        self.application_started = False
        self.service = service
        self.station = station
        self.timers = ConnectionTimers()
        self.frame_state_tracker = FrameStateTracker()
        self.status = ConnectionStatus.DISCONNECTED
        self._settings = self.station.connection.settings
        if settings:
            self._settings.update(settings)
        if frame_sender is None:

            def frame_sender(frame: Frame) -> None:
                self.station.frame_router.send_frame(self.interface, frame)

        self.frame_sender = frame_sender
        self.close_callback = close_callback

    @property
    def settings(self) -> ConnectionSettings:
        """
        Get the tunable settings for this connection.
        """
        return smart_clone(self._settings)

    async def reload_settings(self, settings: ConnectionSettings) -> None:
        """
        Reload the tunable settings for this connection. This will take effect
        the next time the timers are triggered.
        """
        self._settings.update(settings)

    def disconnect(self, immediate: bool = False) -> None:
        """
        Initializes disconnection with the remote server.

        If immediate is True, does not wait for the remote system's response before
        shutting down the connection on our end.
        """
        if self.status in (
            ConnectionStatus.DISCONNECTED,
            ConnectionStatus.AWAIT_DISCONNECTION,
        ):
            return
        if self.frame_state_tracker.received_without_acknowledgement:
            # Need to acknowledge outstanding frames...
            self._send_reception_status(status=SFrameType.RECEIVE_NOT_READY)
        self.status = ConnectionStatus.AWAIT_DISCONNECTION
        # ...And then immediately mark for disconnection.
        self.frame_sender(
            command(
                Frame(
                    route=self.route_to_other(),
                    control=Unnumbered(frame_type=UFrameType.DISCONNECT),
                    pid=None,
                )
            ),
        )
        if self.application and self.application_started:
            self.application.on_disconnect(self)
        if immediate:
            self._close()
            return
        delay_action(
            timer=self.timers.t1,
            action=self._close,
            delay=digipeater_factor(
                interval=self._settings["retry_interval"],
                hops=len(self.digipeaters),
            ),
        )

    def route_to_other(self) -> Route:
        """
        Route to the other station. Determines the 'other' station based on the
        'inbound' flag.
        """
        if self.inbound:
            src = self.second_party
            dest = self.first_party
            digipeaters = tuple(reversed(self.digipeaters))
        else:
            src = self.first_party
            dest = self.second_party
            digipeaters = self.digipeaters
        return Route(
            src=AddressHeader(address=src),
            dest=AddressHeader(address=dest),
            digipeaters=tuple(
                AddressHeader(address=digipeater) for digipeater in digipeaters
            ),
        )

    def _close(self) -> None:
        """
        Shuts down the attached application and performs any needed cleanup.

        If trying to close this connection from an application or service, you will want
        to use the public `disconnect` method.
        """
        if self.status == ConnectionStatus.DISCONNECTED:
            # We're already closed, nothing to do.
            return
        self.status = ConnectionStatus.DISCONNECTED
        if self.application:
            self.application.on_killed(self)
        for field_instance in fields(self.timers):
            getattr(self.timers, field_instance.name).kill()
        self.close_callback(self)

    def _send_accept_frame(self) -> None:
        """
        Send a connection accepted frame.
        """
        self.frame_sender(
            response(
                Frame(
                    route=Route(
                        src=AddressHeader(address=self.second_party),
                        dest=AddressHeader(address=self.first_party),
                        digipeaters=tuple(
                            AddressHeader(address=digipeater)
                            for digipeater in reversed(self.digipeaters)
                        ),
                    ),
                    pid=None,
                    control=Unnumbered(frame_type=UFrameType.UNNUMBERED_ACKNOWLEDGE),
                )
            ),
        )

    def _send_connection_request(self) -> None:
        """
        Send a connection frame to initiate the remote connection and start listening
        for acknowledgement.
        """
        self.frame_sender(
            command(
                Frame(
                    route=self.route_to_other(),
                    control=Unnumbered(frame_type=UFrameType.SET_ASYNC_BALANCED_MODE),
                    pid=None,
                )
            ),
        )

    def _start_app(self) -> None:
        """
        Starts the app if it hasn't already been started.
        """
        if self.application and not self.application_started:
            self.application_started = True
            self.application.on_connect(self)

    def _accept_connection(self) -> None:
        """
        Update the status and send the acceptance frame.
        """
        self.status = ConnectionStatus.CONNECTED
        self._send_accept_frame()
        self._idle_check()
        self._start_app()

    @property
    def match_key(self) -> str:
        """
        Suggested base key string used for registering/unregistering matchers.
        """
        path = Route(
            src=AddressHeader(address=self.first_party),
            dest=AddressHeader(address=self.second_party),
            digipeaters=tuple(
                AddressHeader(address=digipeater) for digipeater in self.digipeaters
            ),
        )
        return f"connection-{self.interface.name}-{path}"

    @property
    def key(self) -> ConnectionKey:
        """
        Returns the connection key for this connection.
        """
        return connection_key(self.first_party, self.second_party, self.interface)

    def _handle_u_frame(self, frame: Frame) -> None:
        """
        Handle unnumbered frames.

        Some commands are not implemented because they are not part of v2.0. Some day
        we may support v2.2, but no one supports it other than Direwolf, so it's not a
        priority.

        Unnumbered Information frames are altogether ignored here. It might be possible
        to send some 'out of band' metadata with them, but this is not supported right
        now, and UI frames are not officially part of a virtual circuit.
        Other services may make use of UI, such as for APRS or similar functions.
        """
        control = cast(Unnumbered, frame.control)
        match control.frame_type:
            case UFrameType.UNNUMBERED_ACKNOWLEDGE:
                if self.status == ConnectionStatus.AWAIT_CONNECTION:
                    self.mark_connected()
                    self.send_next_frames()
                elif self.status == ConnectionStatus.AWAIT_DISCONNECTION:
                    self._close()
                # If we're not opening or closing a connection, we should ignore the
                # frame.
            case UFrameType.SET_ASYNC_BALANCED_MODE:
                if self.status == ConnectionStatus.CONNECTED:
                    # Remote station must not have gotten our initial acceptance,
                    # or this was a forcibly created internal connection.
                    self._send_accept_frame()
                    self._start_app()

            case UFrameType.TEST:
                # Spec says we should echo contents of a test frame.
                frame = response_frame(frame)._replace(info=frame.info)
                self.frame_sender(frame)
            case UFrameType.DISCONNECT:
                frame = response(
                    Frame(
                        route=self.route_to_other(),
                        control=Unnumbered(
                            frame_type=UFrameType.UNNUMBERED_ACKNOWLEDGE
                        ),
                        pid=None,
                    )
                )
                self.frame_sender(frame)
                self._close()
            case UFrameType.DISCONNECT_MODE:
                # Other station seems confused we're contacting it at all, or is
                # otherwise too busy to deal with us. Close out without sending
                # acknowledgement.
                self._close()

    def mark_connected(self) -> None:
        """
        Mark this connection as connected. Use this if we've gotten an Unnumbered
        Acknowledgment, or we can infer one would have been sent due to getting an
        I-frame.
        """
        self.status = ConnectionStatus.CONNECTED
        self.timers.t1.cancel()
        if self.application:
            self.application.on_connect(self)

    def _request_retransmission(self) -> None:
        """
        Requests retransmission of the frame we're next expecting.
        """
        if self.frame_state_tracker.request_retransmit:
            # We're already in the process of requesting a retransmission.
            return
        self.frame_state_tracker.request_retransmit = True
        self.frame_sender(
            Frame(
                route=self.route_to_other(),
                control=Supervisory(
                    frame_type=SFrameType.REJECT_FRAME,
                    receiving_sequence_number=self.frame_state_tracker.vr.value,
                ),
                pid=None,
            ),
        )
        self.frame_state_tracker.received_without_acknowledgement = 0

    def resend_from(self, sequence_number: int) -> None:
        """
        Resend all frames starting from a given sequence number.
        """
        if self.frame_state_tracker.other_station_busy:
            return
        # Make sure these frames are resent with the current variables.
        modifier = build_receive_modifier(self.frame_state_tracker.vr.value)
        for index in self.frame_state_tracker.vs.retrace_from(sequence_number):
            frame: Frame | None = self.frame_state_tracker.outstanding_frames.get(
                index, None
            )
            if frame is None:
                # We've retraced all frames.
                return
            self.frame_sender(modifier(frame))

    def _handle_s_frame(self, frame: Frame) -> None:
        """
        Handle supervisory frames.
        """
        control = cast(Supervisory, frame.control)
        match control.frame_type:
            case SFrameType.RECEIVE_READY:
                if is_command(frame):
                    self._send_reception_status(modifier=response)
                else:
                    self._clear_acknowledged_frames(control.receiving_sequence_number)
                    self.frame_state_tracker.other_station_busy = False
                    self.send_next_frames()
            case SFrameType.REJECT_FRAME:
                self._clear_acknowledged_frames(
                    control.receiving_sequence_number, retry=False
                )
                self.resend_from(control.receiving_sequence_number)
            case SFrameType.RECEIVE_NOT_READY:
                self._clear_acknowledged_frames(
                    control.receiving_sequence_number, retry=False
                )
                self.frame_state_tracker.other_station_busy = True

    def _send_reception_status(
        self,
        status: Literal[
            SFrameType.RECEIVE_READY, SFrameType.RECEIVE_NOT_READY
        ] = SFrameType.RECEIVE_READY,
        modifier: Callable[[Frame], Frame] = lambda x: x,
    ) -> None:
        """
        Sends a 'Receive Ready' or 'Receive Not Ready' frame to inform the remote
        station that we've received their frames up to a specific sequence number.
        We normally don't want to do this unless there are no I-frames we could
        shoehorn this info into.
        """
        frame = Frame(
            route=self.route_to_other(),
            control=Supervisory(
                frame_type=status,
                receiving_sequence_number=self.frame_state_tracker.vr.value,
            ),
            pid=None,
        )
        frame = modifier(frame)
        self.frame_sender(
            frame,
        )
        self.frame_state_tracker.received_without_acknowledgement = 0

    def _handle_i_frame(self, frame: Frame) -> None:
        """
        Handle informational frames.
        """
        control = cast(Info, frame.control)
        if self.status in (
            ConnectionStatus.AWAIT_DISCONNECTION,
            ConnectionStatus.DISCONNECTED,
        ):
            # Ignore newly sent I-frames.
            return
        if self.status == ConnectionStatus.AWAIT_CONNECTION:
            # We can infer accepted status in this case.
            self.mark_connected()
        if control.sending_sequence_number != self.frame_state_tracker.vr.value:
            # Don't use this frame. Instead, request retransmission from remote station.
            self._request_retransmission()
            return
        self.frame_state_tracker.request_retransmit = False
        self.frame_state_tracker.vr.increment()
        self.frame_state_tracker.received_without_acknowledgement += 1
        if self.frame_state_tracker.received_without_acknowledgement >= AX25_SEQ_MAX:
            self._send_reception_status()
        self._clear_acknowledged_frames(control.receiving_sequence_number)
        self.send_next_frames()
        self.timers.t2.cancel()
        if self.application and self.status == ConnectionStatus.CONNECTED:
            try:
                self.application.on_bytes(self, frame.info)
            except Exception as err:
                logger.exception(err)
                self.disconnect()
        self.station.collect_task(
            delay_action(
                timer=self.timers.t2,
                action=self._volunteer_reception_status,
                delay=self._settings["reception_status_delay"],
            )
        )

    def _volunteer_reception_status(self) -> None:
        """
        Sent shortly after we've received some frames to make sure the remote knows
        we've gotten them. Also kicks off the idle counter.
        """
        self._send_reception_status()
        self._idle_check()

    @property
    def idle(self) -> bool:
        """
        Returns True if we have no unacknowledged frames, and we have no frames to send.
        """
        if self.frame_state_tracker.received_without_acknowledgement:
            return False
        return not any(self.frame_state_tracker.outstanding_frames.values())

    def _clear_acknowledged_frames(
        self, sequence_number: int, retry: bool = True
    ) -> None:
        """
        Clear all acknowledged frames from tracking, opening up slots for new
        transmitted frames.
        """
        sequence_number = self.frame_state_tracker.va.before(sequence_number)
        for i in self.frame_state_tracker.va.iterate_up_to(sequence_number):
            self.frame_state_tracker.outstanding_frames[i] = None
        if retry:
            self._retry_outstanding_loop()

    def resend_outstanding(self) -> None:
        """
        Resend all frames that have not been acknowledged.
        """
        self.resend_from(self.frame_state_tracker.va.next())

    def _retry_outstanding_loop(self) -> None:
        """
        Starts the loop for retrying any outstanding frame. Run this any time frames
        are added or the acknowledgement variable is changed.
        """
        retry_loop(
            timer=self.timers.t1,
            action=self.resend_outstanding,
            check=lambda: not any(self.frame_state_tracker.outstanding_frames.values()),
            fail_action=self.disconnect,
            immediate=False,
            retries=self._settings["retries"],
            interval=digipeater_factor(
                interval=self._settings["retry_interval"],
                hops=len(self.digipeaters),
            ),
        )

    def send_next_frames(self) -> None:
        """
        Sends outstanding bytes until our frame slots are consumed, or we run out of
        bytes to send.
        """
        if self.status != ConnectionStatus.CONNECTED:
            # Link is not yet ready.
            return
        if self.frame_state_tracker.other_station_busy:
            # Other station is not ready-- do not send any new information frames.
            return
        frame_sent = False
        base_frame = Frame(
            route=self.route_to_other(),
            control=Info(),
            pid=AX25_PID_TEXT,
            info=b"",
        )
        info_length = self.frame_state_tracker.maximum_transmission_unit - len(
            base_frame
        )
        assert info_length > 0, (
            "Maximum transmission unit is too small to handle even the headers!"
        )
        while (
            self.frame_state_tracker.outstanding_frames.get(
                self.frame_state_tracker.vs.value, None
            )
            is None
        ):
            # Continue sending frames until we've used up all our frame slots, or until
            # we're out of bytes to send.

            # We can cancel the Receive Ready timer since we'll send the update as part
            # of the I-frame control.
            self.timers.t2.cancel()
            byte_string = self.send_buffer[: info_length + 1]
            del self.send_buffer[: info_length + 1]
            if not byte_string:
                # We've sent everything already.
                break
            frame = base_frame._replace(
                control=Info(
                    receiving_sequence_number=self.frame_state_tracker.vr.value,
                    sending_sequence_number=self.frame_state_tracker.vs.value,
                ),
                info=bytes(byte_string),
            )
            self.frame_state_tracker.outstanding_frames[
                self.frame_state_tracker.vs.value
            ] = frame
            self.frame_sender(frame)
            self.frame_state_tracker.vs.increment()
            self.frame_state_tracker.received_without_acknowledgement = 0
            frame_sent = True
        if frame_sent:
            self.timers.t3.cancel()
            self._retry_outstanding_loop()
        self._idle_check()

    def _idle_check(self) -> None:
        """
        kicks off the idle timer.
        """
        if not self.idle:
            return
        self.station.collect_task(
            delay_action(
                timer=self.timers.t3,
                action=self._check_remote_status,
                delay=self._settings["connection_check_interval"],
            )
        )

    @property
    def retry_interval(self) -> int:
        """
        Calculates the retry interval for tasks that await acknowledgement.
        """
        return digipeater_factor(
            interval=self._settings["retry_interval"],
            hops=len(self.digipeaters),
        )

    def _check_remote_status(self) -> None:
        """
        Used when as far as we know nothing's happening and nothing's outstanding.
        Checks with the remote station to see what their status is, in case we're wrong,
        and they've lost contact with us.
        """
        retry_loop(
            timer=self.timers.t1,
            retries=self._settings["retries"],
            action=lambda: self._send_reception_status(modifier=command),
            fail_action=lambda: self.disconnect(immediate=True),
            interval=self.retry_interval,
        )

    def send_bytes(self, bytes_to_send: bytes) -> None:
        """
        Applications can send bytes to connected clients using this function.
        """
        self.send_buffer.extend(bytes_to_send)
        self.send_next_frames()

    def inbound_frame(self, frame: Frame, _interface: "Interface") -> None:
        """
        Handles any inbound frames matching our route path.
        """
        # No longer idle-- we received a frame!
        self.timers.t3.cancel()
        match frame.control:
            case Unnumbered():
                self._handle_u_frame(frame)
            case Supervisory():
                self._handle_s_frame(frame)
            case Info():
                self._handle_i_frame(frame)
        # But we might be newly idle now.
        self._idle_check()

    def frame_matcher(self) -> Matcher:
        """
        Matches received frames.
        """
        if self.inbound:
            return check_all(
                has_src_address(self.first_party),
                has_dest_address(self.second_party),
                has_these_digipeaters(self.digipeaters),
                repeats_completed,
            )
        return check_all(
            has_src_address(self.second_party),
            has_dest_address(self.first_party),
            has_these_digipeaters(tuple(reversed(self.digipeaters))),
            repeats_completed,
        )

    def negotiate(self) -> None:
        """
        Performs connection negotiation, usually called after the connection object
        is initialized.
        """
        if self.inbound:
            self._accept_connection()
        else:
            self.status = ConnectionStatus.AWAIT_CONNECTION
            retry_loop(
                timer=self.timers.t1,
                action=self._send_connection_request,
                fail_action=self._close,
                retries=self._settings["retries"],
                interval=self.retry_interval,
            )

    def shutdown(self) -> None:
        """
        Forcibly shut down the connection. Attempt to inform the remote system, but
        don't wait for confirmation.

        We use this when shutting down the station.
        """
        self.disconnect(immediate=True)


def connection_key(
    party1: Address,
    party2: Address,
    interface: "Interface",
) -> ConnectionKey:
    """
    Key for use in connection tables.
    """
    return cast(
        # sorted eats the information about the number of ints in the tuple,
        # so we have to manually cast here. Please update the cast if additional
        # entries are required.
        ConnectionKey,
        tuple(sorted((hash(party1), hash(party2), hash(interface)))),
    )


@dataclass
class ConnectionTimers:
    """
    A set of timers used by a connection for various purposes.
    """

    t1: Timer = field(default_factory=lambda: Timer(name="T1[Acknowledgement]"))
    t2: Timer = field(default_factory=lambda: Timer(name="T2[Response Delay]"))
    t3: Timer = field(default_factory=lambda: Timer(name="T3[Inactive Link]"))



"""
Constants used for analyzing/parsing/assembling/disassembling AX.25 frames, including
Enums.
"""

import sys
from enum import UNIQUE, Enum, verify

from pax25.utils import EnumReprMixin

# Platform specific
AX25_ENDIAN = sys.byteorder
AX25_CODEPAGE = "utf-8"

# Protocol limits
AX25_MIN_FRAME_LEN = 15  # per AX.25 spec (136 bits w/flags, we don't TX/RX flags)
AX25_DEFAULT_INFOMTU = 256  # per AX.25 spec
AX25_SEQ_MAX = 7  # we only do MOD8 connections
AX25_REPEATER_MAX = 8  # Maximum number of repeaters permitted.

# address bitworks
AX25_ADDR_SIZE = 7
AX25_SSID_MASK_RESERVED = 0b01100000
AX25_SSID_SHIFT_RESERVED = 5
AX25_SSID_MASK_COMMAND = 0b10000000
AX25_SSID_MASK_SSID = 0b00011110
AX25_SSID_MASK_LAST_ADDRESS = 0b00000001
AX25_SSID_SHIFT_SSID = 1

# control bitworks. Python will interpret these as integers, but they are written as
# bytes here for clarity when working with the protocol.
AX25_CTRL_I = 0b00000000
AX25_CTRL_MASK_I = 0b00000001
AX25_CTRL_MASK_RECVSEQ = 0b11100000
AX25_CTRL_INVMASK_RECVSEQ = 0b00011111
AX25_CTRL_SHIFT_RECVSEQ = 5
AX25_CTRL_MASK_SENDSEQ = 0b00001110
AX25_CTRL_SHIFT_SENDSEQ = 1
AX25_SUPERVISOR_MASK_SUBTYPE = 0b00001111
AX25_CTRL_MASK_S = 0b00000011
AX25_CTRL_S = 0b00000001
AX25_CTRL_MASK_UI = 0b00000011
AX25_CTRL_UI = 0b00000011
AX25_CTRL_MASK_PF = 0b00010000
AX25_CTRL_INVMASK_PF = 0b11101111
AX25_CTRL_SHIFT_PF = 4


@verify(UNIQUE)
class FrameType(Enum):
    """
    Enum for the different frame types.
    """

    UNNUMBERED = 0b00000011
    SUPERVISORY = 0b00000001
    INFORMATIONAL = 0b00000000


@verify(UNIQUE)
class UFrameType(EnumReprMixin, Enum):
    """
    Different commands which Unnumbered frames may represent.
    """

    SET_ASYNC_BALANCED_MODE_EXTENDED = 0b01101111
    SET_ASYNC_BALANCED_MODE = 0b00101111
    DISCONNECT = 0b01000011
    DISCONNECT_MODE = 0b00001111
    UNNUMBERED_ACKNOWLEDGE = 0b01100011
    FRAME_REJECT = 0b10000111
    UNNUMBERED_INFORMATION = 0b00000011
    EXCHANGE_IDENTIFICATION = 0b10101111
    TEST = 0b11100011


UCOMMAND_DISPLAY: dict[UFrameType, str] = {
    UFrameType.SET_ASYNC_BALANCED_MODE_EXTENDED: "SABME",
    UFrameType.SET_ASYNC_BALANCED_MODE: "SABM",
    UFrameType.DISCONNECT: "DISC",
    UFrameType.DISCONNECT_MODE: "DM",
    UFrameType.UNNUMBERED_ACKNOWLEDGE: "UA",
    UFrameType.FRAME_REJECT: "FRMR",
    UFrameType.UNNUMBERED_INFORMATION: "UI",
    UFrameType.EXCHANGE_IDENTIFICATION: "XID",
    UFrameType.TEST: "TEST",
}


@verify(UNIQUE)
class SFrameType(EnumReprMixin, Enum):
    """
    Supervisory Frame types.
    """

    RECEIVE_READY = 0b00000001
    RECEIVE_NOT_READY = 0b00000101
    REJECT_FRAME = 0b00001001
    SELECTIVE_REJECT = 0b00001101


SFRAME_DISPLAY: dict[SFrameType, str] = {
    SFrameType.RECEIVE_READY: "RR",
    SFrameType.RECEIVE_NOT_READY: "RNR",
    SFrameType.REJECT_FRAME: "REJ",
    SFrameType.SELECTIVE_REJECT: "SREJ",
}


S_FRAME_MAP: dict[int, SFrameType] = {entry.value: entry for entry in SFrameType}
U_FRAME_MAP: dict[int, UFrameType] = {entry.value: entry for entry in UFrameType}

# Protocol ID types. This one is 'plain text' which almost everything uses.
AX25_PID_TEXT = 0xF0

# Status and commands

AX25_STATUS_DISC = 0  # DISC - DISCONNECTED
AX25_STATUS_CONN = 1  # CONN - CONNECTED
AX25_STATUS_DREQ = 2  # DREQ - DISCONNECT_INBOUND REQUESTED
AX25_STATUS_CREQ = 3  # CREQ - CONNECT_INBOUND REQUESTED
AX25_STATUS_DISP = 4  # DISP - DISPOSE
AX25_STATUS_ERR = 5  # ERRR - ERROR
AX25_STATUS_PTX = 6  # P-TX - PAUSED TX
AX25_STATUS_PRX = 7  # P-RX - PAUSED RX
AX25_STATUS_PRTX = 8  # PRTX - PAUSED TX AND RX

AX25_CMD_CONN = 0  # ATTEMPT TO CONNECT_INBOUND THIS DATALINK
AX25_CMD_DISC = 0  # DISCONNECT_INBOUND THIS DATALINK



"""
Dummy interface. No configuration, ignores everything. Used for tests.
"""

import asyncio
from asyncio import Queue, Task
from typing import TYPE_CHECKING

from ..ax25.frame import Frame
from ..utils import cancel
from .types import DummySettings, Interface

if TYPE_CHECKING:  # pragma: no cover
    from ..station import Station


class DummyInterface(Interface[DummySettings]):
    """
    Dummy interface for testing. Allows you to check the precise frames received and
    send exact frames as needed.
    """

    type = "Dummy"

    def __init__(self, name: str, settings: DummySettings, station: "Station"):
        """
        Just stash the args but don't do anything with them.
        """
        self.name = name
        self._settings = settings
        self.station = station
        self.sent_frames: Queue[Frame] = Queue()
        self.queue: Queue[Frame] = Queue()
        self._loop: Task[None] | None = None

    @property
    def listening(self) -> bool:
        """
        Returns a bool indicating whether the interface is listening.
        """
        if not self._loop:
            return False
        return not self._loop.done()

    @property
    def sudo(self) -> bool:
        """
        Returns if connectsion on this interface should be considered privileged.
        """
        return self._settings.get("sudo", False)

    async def reload_settings(self, settings: DummySettings) -> None:
        """
        Reloads the interface with revised settings.
        """
        self._settings = settings
        await self.shutdown()
        self.start()

    @property
    def gateway(self) -> bool:
        """
        Returns true if we can perform outbound connections via this interface.
        """
        return self._settings.get("gateway", False)

    def send_frame(self, frame: Frame) -> None:
        """
        Dummy send frame function.
        """
        self.sent_frames.put_nowait(frame)

    def start(self) -> None:
        """
        Starts the read loop.
        """
        self.queue = Queue()
        self._loop = asyncio.ensure_future(self.read_loop())

    async def read_loop(self) -> None:
        """
        Dummy read loop function. Probably won't be an issue if it just returns
        immediately.
        """
        while True:
            frame = await self.queue.get()
            self.station.frame_router.process_frame(self, frame)
            self.queue.task_done()

    async def shutdown(self) -> None:
        """
        Dummy shut down function.
        """
        self.queue.shutdown()
        if self._loop:
            await cancel(self._loop)
            self._loop = None



"""
This module contains the different interfaces which are included with pax25.
"""

from .dummy import DummyInterface
from .file import FileInterface
from .serial import SerialInterface
from .tcp import TCPInterface
from .types import Interface

INTERFACE_TYPES = {
    "file": FileInterface,
    "dummy": DummyInterface,
    "serial": SerialInterface,
    "tcp": TCPInterface,
}

__all__ = [
    "FileInterface",
    "Interface",
    "SerialInterface",
    "DummyInterface",
    "TCPInterface",
    "INTERFACE_TYPES",
]



"""
Data classes and helpful for assembling/disassembling an AX.25 frame at large.
"""

import os
from ast import literal_eval
from dataclasses import dataclass
from datetime import datetime
from typing import NamedTuple, Self, cast

from pax25.ax25.address import Route
from pax25.ax25.constants import AX25_ENDIAN
from pax25.ax25.control import (
    Info,
    Supervisory,
    Unnumbered,
    control_from_json,
    derive_control_class,
)
from pax25.ax25.exceptions import DisassemblyError
from pax25.ax25.protocols import Assembler
from pax25.ax25.utils import is_command, is_response, size
from pax25.protocols import JSONObj
from pax25.utils import normalize_line_endings


class Frame(NamedTuple):
    """
    Represents an AX.25 frame for transmission or reception
    """

    pid: int | None
    route: Route
    control: Unnumbered | Supervisory | Info
    info: bytes = b""

    def size(self) -> int:
        return sum(
            size(item or "")
            for item in (
                self.route,
                self.control,
                (self.pid is not None) and self.pid.to_bytes(1, AX25_ENDIAN),
                self.info,
            )
        )

    def __str__(self) -> str:
        """
        String representation of an AX25 frame. Emulates (mostly) how a TNC displays a
        frame, with the main exception being that we display binary data as its hex
        representation rather than sending it literally.
        """
        segments = [
            str(self.route),
            ": ",
        ]
        control_segment = str(self.control)
        if is_command(self):
            control_segment = f"<{control_segment}>"
        elif is_response(self):
            control_segment = f"<{control_segment}>".lower()
        segments.append(control_segment)
        if self.info:
            segments.extend(
                [
                    ":",
                    os.linesep,
                    normalize_line_endings(self.info).decode(
                        encoding="utf-8", errors="backslashreplace"
                    ),
                ]
            )
        return "".join(segments)

    def assemble(self) -> bytes:
        """
        Assemble this frame into a bytearray suitable for transmission.
        """
        data = bytearray()
        data.extend(self.route.assemble())
        data.extend(self.control.assemble())
        # PID could only be set if the control byte is set, per disassembly.
        if self.pid is not None:
            data.extend(bytearray(self.pid.to_bytes(1, AX25_ENDIAN)))
        data.extend(self.info)
        return bytes(data)

    @classmethod
    def disassemble(cls, data: bytes) -> Self:
        """
        Given a bytestream frame pulled from the wire, create an Frame instance.
        """
        data, route = consume_assembler(data, Route)
        control_class = derive_control_class(data)
        data, control = consume_assembler(data, control_class)
        try:
            pid = data[0]
        except IndexError as err:
            if not isinstance(control, Supervisory | Unnumbered):
                raise DisassemblyError("Protocol ID is missing.") from err
            pid = None
        data = data[1:]
        info = bytes(data)
        return cls(
            route=route,
            control=cast(Unnumbered | Info | Supervisory, control),
            pid=pid,
            info=info,
        )

    def to_json(self) -> JSONObj:
        return {
            "__class__": self.__class__.__name__,
            "pid": self.pid,
            "route": self.route.to_json(),
            "control": self.control.to_json(),
            "info": repr(self.info),
        }

    @classmethod
    def from_json(cls, obj: JSONObj) -> Self:
        kwargs = {
            "pid": cast(int, obj["pid"]),
            "route": Route.from_json(cast(JSONObj, obj["route"])),
            "control": control_from_json(cast(JSONObj, obj["control"])),
            "info": literal_eval(cast(str, obj["info"])),
        }
        return cls(**kwargs)


@dataclass
class AxFrameTracker:
    """
    Metadata about a frame, unrelated to its protocol contents.
    """

    frame: Frame | None = None
    tx_time: datetime | None = None
    tx_count: int = 0
    # phy: PhysicalInterface, or similar, when ready.
    fault: bool = False


def consume_assembler[T: Assembler](data: bytes, cls: type[T]) -> tuple[bytes, T]:
    """
    Given a bytearray, pull what's necessary from the array to form a given disassembled
    dataclass, and return it with the remainder.
    """
    instance = cls.disassemble(data)
    return data[instance.size() :], instance



"""
Serial KISS reader.
"""

import asyncio
import logging
import os
import threading
from asyncio import Queue, QueueShutDown, Task
from collections.abc import Callable, Coroutine
from typing import TYPE_CHECKING

from serial import Serial

from pax25.ax25.frame import Frame
from pax25.interfaces.kiss.constants import (
    KISS_CMD_DATA,
    KISS_CMD_FULLDUP,
    KISS_CMD_PERSIST,
    KISS_CMD_SLOTTIME,
    KISS_CMD_TXDELAY,
    KISS_ENDIAN,
)
from pax25.interfaces.kiss.protocol import ax25_frames_from_kiss, kiss_command
from pax25.interfaces.types import (
    Interface,
    SerialSettings,
    TerminalNodeControllerSettings,
)
from pax25.utils import async_wrap, cancel_all, smart_clone

if TYPE_CHECKING:  # pragma: no cover
    from pax25.station import Station


logger = logging.getLogger(__name__)

aprint = async_wrap(print)


def default_tnc_settings() -> TerminalNodeControllerSettings:
    """
    Default settings to send to the TNC.
    """
    return {
        "port": 0,
        "tx_delay": 12,
        "persist": 63,
        "slot_time": 1,
        "full_duplex": 0,
    }


def build_serial_reader(serial: Serial) -> Callable[[], Coroutine[None, None, bytes]]:
    """
    Creates a thread and makes a reader based on it.
    """
    queue = Queue[bytes]()

    def read_loop() -> None:
        while True:
            try:
                data = serial.read(serial.in_waiting or 1)
            except TypeError:
                # Can happen if the file handle is closed abruptly.
                queue.shutdown()
                return
            if not data:
                queue.shutdown()
                return
            queue.put_nowait(data)

    thread = threading.Thread(target=read_loop)
    thread.start()

    buffer = bytearray()

    async def read_next() -> bytes:
        nonlocal buffer
        if not buffer:
            try:
                buffer.extend(await queue.get())
            except QueueShutDown:
                thread.join()
                return b""
        return buffer.pop(0).to_bytes(1, KISS_ENDIAN)

    return read_next


class SerialInterface(Interface[SerialSettings]):
    """
    Interface for KISS serial interaction.
    """

    type = "Serial"
    serial: Serial

    def __init__(self, name: str, settings: SerialSettings, station: "Station") -> None:
        """
        Initialize a serial interface.
        """
        self.name = name
        self.station = station
        self.send_queue: Queue[Frame] = Queue()
        self._settings = settings
        tnc_settings = default_tnc_settings()
        tnc_settings.update(settings.get("tnc_settings", {}))
        self._tnc_settings = tnc_settings
        self._read_loop: Task[None] | None = None
        self._write_loop: Task[None] | None = None
        self.serial = self.initialize_serial()

    def initialize_serial(self) -> Serial:
        """
        Set up the serial interface.
        """
        return Serial(
            self.path,
            baudrate=self.baud_rate,
            bytesize=8,
            parity="N",
            stopbits=1,
            rtscts=self.rtscts,
            dsrdtr=self.dsrdtr,
            xonxoff=self.xonxoff,
        )

    @property
    def settings(self) -> SerialSettings:
        """
        Return the running settings of the serial interface.
        """
        settings = smart_clone(self._settings)
        settings["tnc_settings"] = smart_clone(self._tnc_settings)
        return settings

    @property
    def sudo(self) -> bool:
        """
        Randos over the air shouldn't be superusers.
        """
        return False

    async def reload_settings(self, settings: SerialSettings) -> None:
        """
        Reload the serial interface.
        """
        await self.shutdown()
        self._settings = settings
        tnc_settings = default_tnc_settings()
        tnc_settings.update(settings.get("tnc_settings", {}))
        self._tnc_settings = tnc_settings
        self.serial = self.initialize_serial()
        self.start()

    @property
    def path(self) -> str:
        default_path = "COM1" if os.name == "nt" else "/dev/ttyUSB0"
        return self.settings.get("path", default_path)

    @property
    def baud_rate(self) -> int:
        return self.settings.get("baud_rate", 9600)

    @property
    def timeout(self) -> int:
        return self.settings.get("timeout", 1)

    @property
    def write_timeout(self) -> int:
        return self.settings.get("write_timeout", 2)

    @property
    def rtscts(self) -> bool:
        return self.settings.get("rtscts", False)

    @property
    def dsrdtr(self) -> bool:
        return self.settings.get("dsrdtr", False)

    @property
    def xonxoff(self) -> bool:
        return self.settings.get("xonxoff", False)

    @property
    def exit_kiss(self) -> bool:
        return self.settings.get("exit_kiss", False)

    @property
    def gateway(self) -> bool:
        return self.settings.get("gateway", True)

    @property
    def tx_delay(self) -> bytes:
        return self._tnc_settings["tx_delay"].to_bytes(1, KISS_ENDIAN)

    @property
    def persist(self) -> bytes:
        return self._tnc_settings["persist"].to_bytes(1, KISS_ENDIAN)

    @property
    def slot_time(self) -> bytes:
        return self._tnc_settings["slot_time"].to_bytes(1, KISS_ENDIAN)

    @property
    def full_duplex(self) -> bytes:
        return self._tnc_settings["full_duplex"].to_bytes(1, KISS_ENDIAN)

    @property
    def port(self) -> int:
        return self._tnc_settings["port"]

    @property
    def listening(self) -> bool:
        """
        Returns a bool indicating whether the interface is listening.
        """
        if not self._read_loop:
            return False
        return not self._read_loop.done()

    def start(self) -> None:
        """
        Starts the serial loops.
        """
        if self.exit_kiss:
            self.exit_kiss_mode()
            logger.info("Exit KISS mode invoked. Shutting down immediately.")
            asyncio.ensure_future(self.station.shutdown())
            return
        self._read_loop = asyncio.ensure_future(self.read_loop())
        self._write_loop = asyncio.ensure_future(self.write_loop())

    def send_settings(self) -> None:
        """
        Send the bytes down serial to configure the TNC.
        """
        # Need to write out the bytes to set up the TNC.
        self.serial.write(
            kiss_command(KISS_CMD_TXDELAY, self.port, self.tx_delay)
            + kiss_command(KISS_CMD_PERSIST, self.port, self.persist)
            + kiss_command(KISS_CMD_SLOTTIME, self.port, self.slot_time)
            + kiss_command(KISS_CMD_FULLDUP, self.port, self.full_duplex)
        )

    def exit_kiss_mode(self) -> None:
        """
        Can be used if we need to exit KISS mode, if your TNC supports it.
        """
        self.serial.write(b"\xc0\xff\xc0")

    async def read_loop(self) -> None:
        """
        Connect to the given serial object and then start reading frames.
        """
        async for frame in ax25_frames_from_kiss(build_serial_reader(self.serial)):
            self.station.frame_router.process_frame(self, frame)

    async def write_loop(self) -> None:
        """
        Write loop for serial interface.
        """
        self.send_settings()
        while frame := await self.send_queue.get():
            self.serial.write(kiss_command(KISS_CMD_DATA, self.port, frame.assemble()))

    def send_frame(self, frame: Frame) -> None:
        """
        Send frames out on KISS to the serial device.
        """
        self.send_queue.put_nowait(frame)

    async def shutdown(self) -> None:
        """
        Close the serial connection and stop listening.
        """
        self._read_loop, self._write_loop = await cancel_all(
            [self._read_loop, self._write_loop]
        )
        if self.serial.is_open:
            self.serial.flush()
            self.serial.close()



"""
Quick and common client imports for Pax25.
"""

from . import interfaces
from .applications.application import Application
from .station import Station

__all__ = ["Station", "Application", "interfaces"]



"""
Utility functions for working with AX.25 frames.
"""

from collections.abc import Callable, Sized
from typing import TYPE_CHECKING, TypeVar, cast

from pax25.ax25.address import Address, AddressHeader
from pax25.ax25.control import Info, Supervisory, Unnumbered
from pax25.ax25.protocols import Assembler

if TYPE_CHECKING:  # pragma: no cover
    from pax25.ax25.frame import Frame


def should_repeat_for(address: Address, frame: "Frame") -> bool:
    """
    Checks if a given address is next in the digipeater list and this frame requires
    repeating.
    """
    for header in frame.route.digipeaters:
        if header.command_or_repeated:
            continue
        if header.address == address:
            return True
        # If we're not next in the repeater list, we don't need to repeat.
        break
    return False


def repeated_for(address: Address, frame: "Frame") -> "Frame":
    """
    Returns a revised frame with the repeat flag set for the specific address.
    Raises if this address isn't a repeater for the frame.
    """
    digipeater_list = list(frame.route.digipeaters)
    address_seen = False
    for index, envelope in enumerate(digipeater_list):
        if envelope.address != address:
            continue
        address_seen = True
        if not envelope.command_or_repeated:
            new_envelope = envelope._replace(command_or_repeated=True)
            digipeater_list[index] = new_envelope
            new_route = frame.route._replace(digipeaters=tuple(digipeater_list))
            return frame._replace(route=new_route)
    if address_seen:
        raise ValueError(
            f"{address} is in the digipeater list, "
            f"but all occurrences are marked repeated.",
        )
    raise ValueError(f"{address} is not a digipeater for {frame}.")


def reply_digipeaters(
    digipeaters: tuple[AddressHeader, ...],
) -> tuple[AddressHeader, ...]:
    """
    Returns a revised frame that has the digipeaters in reverse order and the repeated
    flags all set to False.
    """
    revised_digipeaters = [
        envelope._replace(command_or_repeated=False)
        for envelope in reversed(digipeaters)
    ]
    return tuple(revised_digipeaters)


def response_frame(frame: "Frame") -> "Frame":
    """
    Generates the base of a response frame-- source and destination are swapped,
    digipeaters reversed, and contents emptied. The control is unchanged, as context
    for changes to it would be kept elsewhere.

    The command_or_repeated flag is swapped along with the addresses
    (if it is set at all), so bear this in mind.
    """
    route = frame.route
    route = route._replace(
        src=route.dest,
        dest=route.src,
        digipeaters=reply_digipeaters(route.digipeaters),
    )
    return frame._replace(route=route, info=b"")


T = TypeVar("T")


def update_receive_number(frame: "Frame", new_number: int) -> "Frame":
    """
    Updates the receive number, if applicable, on the target frame.
    """
    match frame.control:
        case Supervisory() | Info():
            return frame._replace(
                control=frame.control._replace(receiving_sequence_number=new_number),
            )
        case _:
            # Unnumbered, no need to update.
            return frame


def build_receive_modifier(new_number: int) -> Callable[["Frame"], "Frame"]:
    """
    Creates a modifier function that will replace an information or supervisory frame's
    receive control number with the `new_number` specified.
    """

    def wrapped(frame: "Frame") -> "Frame":
        return update_receive_number(frame, new_number)

    return wrapped


C = TypeVar("C", bound=Supervisory | Unnumbered | Info)


def flagged_control(control: C) -> C:
    """
    Marks the poll/response bit for this frame, or raises if we're an info frame.
    """
    match control:
        case Supervisory():
            control = cast(
                C,
                control._replace(poll_or_final=True),  # type: ignore
            )
        case Unnumbered():
            control = cast(
                C,
                control._replace(poll_or_final=True),  # type: ignore
            )
        case _:
            raise ValueError(
                f"{control.__class__.__name__} control fields cannot be "
                f"marked as command/response."
            )
    return control


def response(frame: "Frame") -> "Frame":
    """
    Mark the frame as a 'Response' frame, replying to a command frame.
    """
    frame = frame._replace(control=flagged_control(frame.control))
    return frame._replace(
        route=frame.route._replace(
            src=frame.route.src._replace(command_or_repeated=True),
            dest=frame.route.dest._replace(command_or_repeated=False),
        ),
    )


def command(frame: "Frame") -> "Frame":
    """
    Mark the frame as a 'Command' frame, sent from the source to the destination.
    """
    frame = frame._replace(control=flagged_control(frame.control))
    return frame._replace(
        route=frame.route._replace(
            src=frame.route.src._replace(command_or_repeated=False),
            dest=frame.route.dest._replace(command_or_repeated=True),
        ),
    )


def is_neutral(frame: "Frame") -> bool:
    """
    Determine whether a frame is neutral (not a command or response frame)
    """
    if frame.route.dest.command_or_repeated == frame.route.src.command_or_repeated:
        return True
    return not getattr(frame.control, "poll_or_final", False)


def is_command(frame: "Frame") -> bool:
    """
    Determine whether the frame is a command frame.
    """
    if is_neutral(frame):
        return False
    return (
        frame.route.dest.command_or_repeated and not frame.route.src.command_or_repeated
    )


def is_response(frame: "Frame") -> bool:
    """
    Determine whether the frame is a response frame.
    """
    if is_neutral(frame):
        return False
    return (
        frame.route.src.command_or_repeated and not frame.route.dest.command_or_repeated
    )


def roll_back_ssid(address: Address) -> Address:
    """
    Given an address, return the previous SSID. This is used by nodes connecting outward
    to prevent loopback issues, by making sure that the SSID on the next connection
    outward is different from the current one, and in a way that will be inaudible by
    the time it loops back around to the original station.
    """
    ssid = address.ssid
    ssid -= 1
    if ssid < 0:
        ssid = 15
    return address._replace(ssid=ssid)


def size(item: Assembler | Sized) -> int:
    """
    Get the size of an item for sending across the wire.
    """
    if hasattr(item, "size"):
        return item.size()
    return len(item)



"""
The contrib module contains premade applications and plugins that we anticipate will be
widely useful both as examples and for practical use to client developers setting up
their own station.

These allow us to eat our own dogfood, and test the breadth of capability of the library
all in one. They are NOT the only applications we expect to develop, nor the only ones
we expect to use as examples, but they are the ones that we use for testing and as
'batteries included' functionality.
"""

from .bulletin_board import BulletinBoard
from .command.command import CommandLine
from .echo import Echo
from .exec import Exec

__all__ = ["BulletinBoard", "CommandLine", "Echo", "Exec"]



"""
Constants for the KISS protocol.
"""

import sys

KISS_FEND = b"\xc0"
KISS_FESC = b"\xdb"
KISS_TFEND = b"\xdc"
KISS_TFESC = b"\xdd"
KISS_ESCAPED_FEND = b"\xdb\xdc"
KISS_ESCAPED_FESC = b"\xdb\xdd"

KISS_MIN_FRAME_LEN = 15

KISS_CMD_DATA = 0x0  # For sending frames.
KISS_CMD_TXDELAY = 0x1  # 1 - TXDELAY (next byte is value in 10ms units (def 50))
KISS_CMD_PERSIST = 0x2  # 2 - PERSIST (next byte is p value in equation (def 63))
KISS_CMD_SLOTTIME = (
    0x3  # 3 - SLOTTIME (next byte is slot interval in 10ms units (def 10))
)
KISS_CMD_CUSTOM4 = 0x4  # 4 - UNUSED - was TXTAIL
KISS_CMD_FULLDUP = 0x5  # 5 - FULLDUPLEX (next byte > 0  FD, 0 HD (def 0))
KISS_CMD_SET_HARDWARE = 0x6  # 6 - SET-HARDWARE - user definable
KISS_CMD_PASSWORD = 0x7  # 7 - We use this for APRS password exchange over TCP, but for
# Some hardware devices over serial, it may have other uses.
KISS_CMD_PASSWORD_REJECT = 0x8  # 8 - Another user-definable value. We use this for the
# password rejected command, but it may have other uses.

KISS_MASK_PORT = 0b11110000
KISS_SHIFT_PORT = 4
KISS_MASK_CMD = 0b00001111

KISS_ENDIAN = sys.byteorder



"""
Control Frame Types
There are three control field types in AX.25, which correspond to the three AX.25 frame
types. These are the Info, Supervisory, and Unnumbered frame types.

Semantics change significantly based on which control type is in use. Therefore, we have
multiple dataclasses for the purpose of handling these different control field types,
as well as helper functions for identifying them.

Supervisory and Informational frames come in two flavors, modulo 8 and modulo 128.
Modulo 128 uses two-byte control fields and allows for a larger sequence frame count.

For the moment, we only support modulo 8, which means we assume all control fields to be
one byte long and consider anything else to be an error. In the future, modulo 128 may
be supported.
"""

from typing import Literal, NamedTuple, Self, cast

from pax25.ax25.constants import (
    AX25_CTRL_I,
    AX25_CTRL_INVMASK_PF,
    AX25_CTRL_MASK_I,
    AX25_CTRL_MASK_PF,
    AX25_CTRL_MASK_RECVSEQ,
    AX25_CTRL_MASK_S,
    AX25_CTRL_MASK_SENDSEQ,
    AX25_CTRL_MASK_UI,
    AX25_CTRL_S,
    AX25_CTRL_SHIFT_PF,
    AX25_CTRL_SHIFT_RECVSEQ,
    AX25_CTRL_SHIFT_SENDSEQ,
    AX25_CTRL_UI,
    AX25_ENDIAN,
    AX25_SUPERVISOR_MASK_SUBTYPE,
    S_FRAME_MAP,
    SFRAME_DISPLAY,
    U_FRAME_MAP,
    UCOMMAND_DISPLAY,
    FrameType,
    SFrameType,
    UFrameType,
)
from pax25.ax25.exceptions import DisassemblyError
from pax25.protocols import JSONObj


class Unnumbered(NamedTuple):
    """
    Unnumbered control header flags.
    """

    frame_type: UFrameType
    poll_or_final: bool = False

    @property
    def type(self) -> Literal[FrameType.UNNUMBERED]:
        """
        Returns the frame's type.
        """
        return FrameType.UNNUMBERED

    def size(self) -> int:
        """
        Unnumbered control is one byte, always.
        """
        return 1

    def __str__(self) -> str:
        """
        String representation of an unnumbered control field.
        """
        return f"<{UCOMMAND_DISPLAY[self.frame_type]}>"

    @classmethod
    def disassemble(cls, data: bytes) -> Self:
        """
        Disassembles a control byte for an unnumbered frame.
        """
        byte = data[0]
        if (byte & AX25_CTRL_UI) != AX25_CTRL_UI:
            raise DisassemblyError(
                f"Attempted to disassemble a control byte as an "
                f"unnumbered control byte, but it was something else. "
                f"Byte was {byte!r}."
            )

        frame_type = U_FRAME_MAP[byte & AX25_CTRL_INVMASK_PF]
        poll_final = bool(byte & AX25_CTRL_MASK_PF)
        return cls(frame_type=frame_type, poll_or_final=poll_final)

    def assemble(self) -> bytes:
        """
        Assembles a control byte for an unnumbered frame.
        """
        data = int(self.poll_or_final)
        data <<= AX25_CTRL_SHIFT_PF
        data |= self.frame_type.value
        return data.to_bytes(1, AX25_ENDIAN)

    @classmethod
    def from_json(cls, obj: JSONObj) -> Self:
        kwargs = {
            "frame_type": getattr(UFrameType, cast(str, obj["frame_type"]).strip("_")),
            "poll_or_final": cast(bool, obj["poll_or_final"]),
        }
        return cls(**kwargs)

    def to_json(self) -> JSONObj:
        return {
            "__class__": self.__class__.__name__,
            "frame_type": self.frame_type.name,
            "poll_or_final": self.poll_or_final,
        }


class Supervisory(NamedTuple):
    """
    Supervisor control header flags.
    """

    frame_type: SFrameType
    receiving_sequence_number: int = 0
    poll_or_final: bool = False

    @property
    def type(self) -> Literal[FrameType.SUPERVISORY]:
        """
        Returns the frame's type.

        Note: In addition to identifying the frame as a supervisory frame, supervisory
        frames have their own subtypes. The subtype is stored on the frame_type
        property.
        """
        return FrameType.SUPERVISORY

    def __str__(self) -> str:
        """
        String representation of a supervisory control field.
        """
        return f"<{SFRAME_DISPLAY[self.frame_type]}>"

    def size(self) -> int:
        """
        Supervisory control is one byte, always.
        """
        return 1

    @classmethod
    def disassemble(cls, data: bytes) -> Self:
        """
        Disassembles a control byte for a supervisory frame.
        """
        byte = data[0]
        if (byte & AX25_CTRL_MASK_S) != FrameType.SUPERVISORY.value:
            raise DisassemblyError(
                f"Attempted to disassemble a non-supervisory control byte as a "
                f"supervisory control byte. Byte was {byte!r}."
            )
        supervisory_frame_type = byte & AX25_SUPERVISOR_MASK_SUBTYPE
        poll_final = bool(byte & AX25_CTRL_MASK_PF)
        receive_sequence_number = byte & AX25_CTRL_MASK_RECVSEQ
        receive_sequence_number >>= AX25_CTRL_SHIFT_RECVSEQ
        return cls(
            receiving_sequence_number=receive_sequence_number,
            poll_or_final=poll_final,
            frame_type=S_FRAME_MAP[supervisory_frame_type],
        )

    def assemble(self) -> bytes:
        """
        Assembles a control byte for an info frame.
        """
        byte = 0
        byte |= self.receiving_sequence_number << AX25_CTRL_SHIFT_RECVSEQ
        byte |= self.poll_or_final << AX25_CTRL_SHIFT_PF
        byte |= self.frame_type.value
        return byte.to_bytes(1, AX25_ENDIAN)

    def to_json(self) -> JSONObj:
        return {
            "__class__": self.__class__.__name__,
            "frame_type": self.frame_type.name,
            "receiving_sequence_number": self.receiving_sequence_number,
            "poll_or_final": self.poll_or_final,
        }

    @classmethod
    def from_json(cls, obj: JSONObj) -> Self:
        kwargs = {
            "frame_type": getattr(SFrameType, cast(str, obj["frame_type"]).strip("_")),
            "receiving_sequence_number": cast(int, obj["receiving_sequence_number"]),
            "poll_or_final": cast(bool, obj["poll_or_final"]),
        }
        return cls(**kwargs)


class Info(NamedTuple):
    """
    Info control header flags.
    """

    sending_sequence_number: int = 0
    receiving_sequence_number: int = 0
    poll_or_final: bool = False

    @property
    def type(self) -> Literal[FrameType.INFORMATIONAL]:
        """
        Returns the bit value of the info frame type identifier.
        """
        return FrameType.INFORMATIONAL

    def __str__(self) -> str:
        """
        String representation of an info control field.
        """
        return "<I>"

    def size(self) -> int:
        """
        Info control is one byte, always.
        """
        return 1

    @classmethod
    def disassemble(cls, data: bytes) -> Self:
        """
        Disassembles a control byte for an info frame.
        """
        byte = data[0]
        if (byte & AX25_CTRL_MASK_I) != FrameType.INFORMATIONAL.value:
            raise DisassemblyError(
                f"Attempted to disassemble a non-info control byte as a control byte. "
                f"Byte was {byte!r}."
            )
        receiving_sequence_number = (
            byte & AX25_CTRL_MASK_RECVSEQ
        ) >> AX25_CTRL_SHIFT_RECVSEQ
        sending_sequence_number = (
            byte & AX25_CTRL_MASK_SENDSEQ
        ) >> AX25_CTRL_SHIFT_SENDSEQ
        poll_final = bool(byte & AX25_CTRL_MASK_PF)
        return cls(
            receiving_sequence_number=receiving_sequence_number,
            sending_sequence_number=sending_sequence_number,
            poll_or_final=poll_final,
        )

    def assemble(self) -> bytes:
        """
        Assembles a control byte for an info frame.
        """
        data = 0
        data |= self.receiving_sequence_number << AX25_CTRL_SHIFT_RECVSEQ
        data |= self.sending_sequence_number << AX25_CTRL_SHIFT_SENDSEQ
        data |= self.poll_or_final << AX25_CTRL_SHIFT_PF
        return data.to_bytes(1, AX25_ENDIAN)

    def to_json(self) -> JSONObj:
        return {
            "__class__": "Info",
            "sending_sequence_number": self.sending_sequence_number,
            "receiving_sequence_number": self.receiving_sequence_number,
            "poll_or_final": self.poll_or_final,
        }

    @classmethod
    def from_json(cls, obj: JSONObj) -> Self:
        kwargs = {
            "sending_sequence_number": cast(int, obj["sending_sequence_number"]),
            "receiving_sequence_number": cast(int, obj["receiving_sequence_number"]),
            "poll_or_final": cast(bool, obj["poll_or_final"]),
        }
        return cls(**kwargs)  # type: ignore[arg-type]


def derive_control_class(
    data: bytes,
) -> type[Info] | type[Unnumbered] | type[Supervisory]:
    """
    Derives the appropriate control class for a given control byte.
    """
    byte = data[0]
    if (byte & AX25_CTRL_MASK_I) == AX25_CTRL_I:
        return Info
    if (byte & AX25_CTRL_MASK_UI) == AX25_CTRL_UI:
        return Unnumbered
    if (byte & AX25_CTRL_MASK_S) == AX25_CTRL_S:
        return Supervisory
    # Pretty sure it's actually impossible to hit this point, since we're using two bits
    # and all options are accounted for.
    raise DisassemblyError(  # pragma: no cover
        f"Could not identify the correct control field type! Byte was {byte!r}"
    )


control_class_dict: dict[str, type[Unnumbered] | type[Info] | type[Supervisory]] = {
    "Unnumbered": Unnumbered,
    "Info": Info,
    "Supervisory": Supervisory,
}


def control_from_json(obj: JSONObj) -> Unnumbered | Info | Supervisory:
    """
    Given a dict representation of a control sequence, return the native control
    structure.
    """
    if obj["__class__"] not in ("Unnumbered", "Info", "Supervisory"):
        raise TypeError("Control class not recognized!")
    return control_class_dict[cast(str, obj["__class__"])].from_json(obj)



"""
Echo application. Echoes back whatever you send to it.
"""

from pax25.applications import Application
from pax25.applications.utils import send_message
from pax25.services.connection.connection import Connection
from pax25.types import EmptyDict


class Echo(Application[EmptyDict]):
    """
    Echo application.
    """

    def on_message(self, connection: "Connection", message: str) -> None:
        """
        Immediately redirect a received message back to the user.
        """
        send_message(connection, message)



"""
Matching utilities. Used for composing routing rules.
"""

from collections.abc import Callable
from typing import NamedTuple, cast

from pax25.ax25.address import Address
from pax25.ax25.constants import FrameType, SFrameType, UFrameType
from pax25.ax25.control import Supervisory, Unnumbered
from pax25.ax25.frame import Frame
from pax25.ax25.utils import should_repeat_for
from pax25.interfaces import Interface

# All matcher functions take a frame and return True for a successful match
# and False if unsuccessful. More complex matching criteria are composed by
# combining simpler matchers.
type Matcher = Callable[[Frame, Interface], bool]
type Notifier = Callable[[Frame, Interface], None]


class MatchCall(NamedTuple):
    """
    A pair of functions-- one for matching, and one for processing a frame.

    Used by the FrameRouter-- if a matcher function matches, calls the notify function.

    Notifiers must be synchronous. If they need to perform an async action, have them
    use an async queue to defer the result. This is to prevent blocking the entire
    station on an async background action.
    """

    matcher: Matcher
    notify: Notifier


def check_all(*checks: Matcher) -> Matcher:
    """
    Returns True if all specified matcher functions return True.

    !!! note

        Passing no matchers will always return True.

    !!! tip

        Put faster-matching functions near the beginning of the list. Matching
        functions which iterate over lists will be used on every frame, potentially
        slowing processing if they are invoked when a faster matcher would otherwise
        decide the matching status.
    """

    def wrapped(frame: Frame, interface: Interface) -> bool:
        """Bound wrapper."""
        return all(check(frame, interface) for check in checks)

    return wrapped


def check_any(*checks: Matcher) -> Matcher:
    """
    Returns True if any of the specified matcher functions return True.

    !!! note

        Passing no matchers will always return True.

    !!! tip

        Put faster-matching functions near the beginning of the list. Matching
        functions which iterate over lists will be used on every frame, potentially
        slowing processing if they are invoked when a faster matcher would otherwise
        decide the matching status.
    """

    def wrapped(frame: Frame, interface: Interface) -> bool:
        """Bound wrapper."""
        return any(check(frame, interface) for check in checks)

    return wrapped


def has_src_address(address: Address) -> Matcher:
    """
    Check if the given frame has the specific destination address.
    """

    def wrapped(frame: Frame, _interface: Interface) -> bool:
        return frame.route.src.address == address

    return wrapped


def has_dest_address(address: Address) -> Matcher:
    """
    Check if the given frame has the specific destination address.
    """

    def wrapped(frame: Frame, _interface: Interface) -> bool:
        """Bound wrapper."""
        return frame.route.dest.address == address

    return wrapped


def needs_repeat_from(address: Address) -> Matcher:
    """
    Checks if the given frame needs repeating by the given address.

    Only does so if it's the next repeater in the set. Otherwise, ignores.
    """

    def wrapped(frame: Frame, _interface: Interface) -> bool:
        """Bound wrapper."""
        return should_repeat_for(address, frame)

    return wrapped


def has_these_digipeaters(digipeaters: tuple[Address, ...]) -> Matcher:
    """
    Checks if a connection has a specific list of intermediate digipeaters.
    """

    def wrapped(frame: Frame, _interface: Interface) -> bool:
        """Bound wrapper."""
        return digipeaters == tuple(
            digipeater.address for digipeater in frame.route.digipeaters
        )

    return wrapped


def is_s_frame(frame: Frame, _interface: Interface) -> bool:
    """
    Checks if the frame is an S frame.
    """
    return frame.control.type == FrameType.SUPERVISORY


def is_u_frame(frame: Frame, _interface: Interface) -> bool:
    """
    Checks if the frame is a U frame.
    """
    return frame.control.type == FrameType.UNNUMBERED


def _u_frame_check(command: UFrameType) -> Matcher:
    """
    Creates a matcher based on a specific UCommand.
    """

    def command_check(frame: Frame, interface: Interface) -> bool:
        """
        Checks if the command type on a frame is a specific ucommand.
        """
        if is_u_frame(frame, interface):
            return cast(Unnumbered, frame.control).frame_type == command
        return False

    return command_check


def _s_frame_check(frame_type: SFrameType) -> Matcher:
    """
    Creates a matcher based on a specific UCommand.
    """

    def frame_type_check(frame: Frame, interface: Interface) -> bool:
        """
        Checks if the command type on a frame is a specific ucommand.
        """
        if is_s_frame(frame, interface):
            return cast(Supervisory, frame.control).frame_type == frame_type
        return False

    return frame_type_check


U_FRAME_CHECKS = {command: _u_frame_check(command) for command in UFrameType}
S_FRAME_CHECKS = {frame_type: _s_frame_check(frame_type) for frame_type in SFrameType}


def is_i_frame(frame: Frame, _interface: Interface) -> bool:
    """
    Checks if the frame is an I-frame.
    """
    return frame.control.type == FrameType.INFORMATIONAL


def repeats_completed(frame: Frame, _interface: Interface) -> bool:
    """
    Checks that a frame has finished making its way through the repeat chain.
    """
    if not frame.route.digipeaters:
        return True
    return all(digi.command_or_repeated for digi in frame.route.digipeaters)


def is_connection_frame_for(
    src: Address, dest: Address, digipeaters: tuple[Address, ...]
) -> Matcher:
    """
    Checks if a frame belongs to a specific connection defined by the src, dest, and
    digipeater path.
    """
    return check_all(
        has_src_address(src),
        has_dest_address(dest),
        has_these_digipeaters(digipeaters),
        repeats_completed,
        check_any(
            is_s_frame,
            is_i_frame,
        ),
    )


def on_gateway(_frame: Frame, interface: Interface) -> bool:
    """
    Checks if a frame was heard on a gateway interface.
    """
    return interface.gateway



import asyncio
import contextlib
import logging
import os
from asyncio import StreamReader, StreamWriter, subprocess
from asyncio.subprocess import Process
from typing import TypedDict

from pax25 import Application
from pax25.services.connection.connection import Connection

logger = logging.getLogger()


class ExecSettings(TypedDict):
    """
    Settings for the exec app.
    """

    path: str
    args: list[str]
    working_directory: str
    environment: dict[str, str]


class ManagedProcess:
    """
    Process manager class that handles starting and communicating with the process, and
    then sends any data to the Exec app.
    """

    stdin: StreamWriter | None
    stdout: StreamReader | None
    proc: Process | None

    def __init__(self, connection: Connection, settings: ExecSettings):
        self.stdin = None
        self.stdout = None
        self.proc = None
        self.prebuffer = b""
        self.connection = connection
        self.settings = settings
        self._initialization = asyncio.ensure_future(self.initialize_process())

    def gen_environment(self) -> dict[str, str]:
        """
        Generate environment variables for the subprocess we'll be creating.
        """
        environment = {
            "PAX25_FIRST_PARTY": str(self.connection.first_party),
            "PAX25_SECOND_PARTY": str(self.connection.second_party),
            "PAX25_INTERFACE_NAME": self.connection.interface.name,
            "PAX25_INTERFACE_TYPE": self.connection.interface.type,
        }
        environment.update(self.settings["environment"])
        return environment

    def close(self) -> None:
        if self.proc:
            with contextlib.suppress(ProcessLookupError):
                self.proc.terminate()

    async def initialize_process(self) -> None:
        """
        Initialize the subprocess that will handle the connection, and start the
        read loop.
        """
        try:
            self.proc = await asyncio.create_subprocess_exec(
                self.settings["path"],
                *self.settings["args"],
                cwd=self.settings["working_directory"] or os.getcwd(),
                stderr=subprocess.STDOUT,
                stdout=asyncio.subprocess.PIPE,
                stdin=asyncio.subprocess.PIPE,
                limit=1,
                env=self.gen_environment(),
            )
            self.stdin, self.stdout = self.proc.stdin, self.proc.stdout
            assert self.stdin
            assert self.stdout
            self.stdin.write(self.prebuffer)
            await self.stdin.drain()
            self.prebuffer = b""
            while self.proc.returncode is None:
                data = await self.stdout.read(1000)
                if data == b"":
                    break
                self.connection.send_bytes(data)
            # Send any remainder.
            self.connection.send_bytes(await self.stdout.read())
        except Exception as err:
            logger.exception(err)
        self.connection.disconnect()

    def send_bytes(self, data: bytes) -> None:
        """
        Write to the external process, or prebuffer if we're still launching.
        """
        if self.stdin is None:
            self.prebuffer += data
            return
        try:
            self.stdin.write(data)
        except OSError:
            self.connection.disconnect()
        except Exception as err:
            logger.exception(err)
            self.connection.disconnect()


class Exec(Application[ExecSettings]):
    """
    The Exec app is a shim application that allows you to make any connection launch
    an executable and redirect its input and output to the connected client.
    """

    connections: dict[Connection, ManagedProcess]
    manager_class = ManagedProcess

    def setup(self) -> None:
        self.connections = {}

    def on_startup(self, connection: "Connection") -> None:
        self.connections[connection] = ManagedProcess(connection, self.settings)

    def on_message(
        self,
        connection: "Connection",
        message: str,
    ) -> None:
        self.connections[connection].send_bytes((message + os.linesep).encode("utf-8"))

    def on_shutdown(self, connection: "Connection") -> None:
        if connection in self.connections:
            self.connections[connection].close()
            del self.connections[connection]



"""
Types used for type checking the library. Client developers can use these if they like,
but they're mostly for our own sanity checks during development/testing.
"""

from typing import TYPE_CHECKING, NotRequired, Protocol, Self, TypedDict

if TYPE_CHECKING:  # pragma: no cover
    from typing import Any

    from pax25.applications import BaseApplication


class EmptyDict(TypedDict):
    """
    Empty dictionary.
    """


class ConnectionSettings(TypedDict):
    """
    Tunable connection parameters. You can probably leave these alone.
    """

    retries: int
    retry_interval: int
    connection_check_interval: int
    reception_status_delay: int


class DigipeaterSettings(TypedDict):
    """
    Settings for the digipeater.
    """

    enabled: NotRequired[bool]


class MonitorSettings(TypedDict):
    """
    Settings for monitoring.
    """

    # How long the frame log is permitted to get. The frame log is typically consumed
    # by an app, such as the contributed CLI Application. If None, tracks forever.
    max_frame_log_size: int | None
    # How many stations to keep track of. Older stations are rolled off. If None, never
    # stops tracking.
    max_stations_tracked: int | None
    # Directory to a file where we can save heard stations between restarts. Empty
    # string (the default) does not save.
    memory_file: str
    # How often to save to the memory file, in seconds. We always save on shutdown,
    # but we also save periodically so we don't lose everything in case of a crash.
    # Default is 60.
    save_interval: int


class BeaconServiceSettings(TypedDict):
    """
    Settings for the beacon service.
    """

    id_beacon_enabled: bool
    # For most jurisdictions, this should be 600 (10 mins) or lower.
    id_beacon_interval: int | float
    id_beacon_destination: str
    id_beacon_digipeaters: dict[str, list[str]]
    id_beacon_content: str | None


# The StationConfig object should always be JSON-serializable. This way we'll eventually
# be able to use JSON for configuration files.


# This means that all interface configs will ALSO need to be JSON-serializable.
class StationConfig(TypedDict):
    """
    Configuration for a station.
    """

    name: str
    digipeater: NotRequired[DigipeaterSettings]
    monitor: NotRequired[MonitorSettings]
    connection: NotRequired[ConnectionSettings]
    beacon: NotRequired[BeaconServiceSettings]


# Map for the station's application registry
ApplicationMap = dict[str, dict[str, "BaseApplication[Any]"]]


class Version(TypedDict):
    """
    Version namedtuple structure. Use this to mark your application versions.
    """

    major: int
    minor: int
    patch: int


class Updatable(Protocol):
    """
    Protocol for objects that can update their values with a copy of themselves,
    I.E., dicts and TypedDicts.
    """

    def update(self, new_values: Self) -> None: ...
